"""
Testing for the DFT-D4 program harness.

Most of the tests use mindless molecules for the diversity of element species
to test as much different interactions as possible.
"""

import numpy as np
import pytest
import qcelemental as qcel

import qcengine as qcng
from qcengine.testing import checkver_and_convert, from_v2, schema_versions, uusing


@uusing("dftd4")
def test_dftd4_task_b97m_m01(schema_versions, request):
    models, retver, _ = schema_versions

    thr = 1.0e-8

    return_result = -0.025024986301735823

    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-01", return_dict=True)),
            specification={
                "model": {"method": "b97m"},
                "driver": "energy",
            },
        )
    else:
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-01", return_dict=True)),
            model={"method": "b97m"},
            driver="energy",
        )

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    atomic_result = qcng.compute(atomic_input, "dftd4", return_version=retver)
    atomic_result = checkver_and_convert(atomic_result, request.node.name, "post")

    print(atomic_result.return_result)
    assert atomic_result.success
    assert pytest.approx(atomic_result.return_result, abs=thr) == return_result


@uusing("dftd4")
def test_dftd4_task_tpss_m02(schema_versions, request):
    models, retver, _ = schema_versions

    thr = 2.0e-8

    return_result = np.array(
        [
            [-1.1006300442501966e-04, -1.9882677409412426e-04, +9.7969330873203453e-05],
            [-9.0564826517969396e-04, -1.1221656220960802e-04, +7.4420608164684399e-04],
            [-8.7360785425552928e-05, +2.8146769568946061e-04, +5.8528720048366145e-04],
            [-1.7602264871464991e-04, -3.0981396137083546e-04, -1.1315671174001885e-03],
            [+6.2854950429890028e-05, +3.1662886959557454e-04, +1.8192682295895164e-04],
            [-1.2033574062747143e-04, -5.8117271931259600e-05, -2.4103707001727414e-04],
            [-9.6727467530178130e-05, -1.2937750080676020e-04, -3.0565838265275271e-05],
            [-1.2139762550383215e-05, +2.2334694600929773e-05, +6.7681766980205114e-05],
            [+2.9131236664176958e-04, -2.3159856865643386e-04, +3.9334087528658355e-04],
            [-8.9512907514913525e-05, +9.3857773376367446e-06, -2.6351331811570630e-04],
            [+3.9621947274011518e-04, -5.8926113817768316e-04, +1.8422719633870202e-04],
            [-1.6466329096860149e-04, +1.3720763821983788e-04, -8.3839852301748649e-05],
            [-3.2173735474065368e-04, +7.2068238657310593e-04, -3.1811972689724591e-04],
            [+2.1298820577529375e-04, -9.6810383911168402e-05, +1.0273702281858908e-04],
            [-4.4599095547508355e-04, -1.8320769460464104e-04, +1.8653129262487302e-04],
            [+1.5668271875651332e-03, +4.2152279374596904e-04, -4.7526466701417510e-04],
        ]
    )

    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-02", return_dict=True)),
            specification={
                "model": {"method": ""},
                "keywords": {
                    "params_tweaks": {
                        "s8": 1.76596355,
                        "a1": 0.42822303,
                        "a2": 4.54257102,
                    }
                },
                "driver": "gradient",
                "extras": {"mymsg": "will I pass through the calc?"},
            },
        )
    else:
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-02", return_dict=True)),
            model={"method": ""},
            keywords={
                "params_tweaks": {
                    "s8": 1.76596355,
                    "a1": 0.42822303,
                    "a2": 4.54257102,
                },
            },
            driver="gradient",
            extras={"mymsg": "will I pass through the calc?"},
        )

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    atomic_result = qcng.compute(atomic_input, "dftd4", return_version=retver)
    atomic_result = checkver_and_convert(atomic_result, request.node.name, "post")

    assert atomic_result.success
    assert pytest.approx(atomic_result.return_result, abs=thr) == return_result
    if "v2" in request.node.name:
        assert "will I pass" in atomic_result.input_data.specification.extras.get(
            "mymsg", "no key!"
        ), "input extras roundtrip fail"
        assert "mymsg" not in atomic_result.extras.get("mymsg", "no key!"), "input extras wrongly present in result"
    else:
        assert "will I pass" in atomic_result.extras.get("mymsg", "no key!"), "input extras roundtrip fail"


@uusing("dftd4")
def test_dftd4_task_r2scan_m03(schema_versions, request):
    models, retver, _ = schema_versions

    thr = 1.0e-8

    return_result = np.array(
        [
            [-1.0184389290110640e-05, -1.2902131714706869e-05, -4.7991389860088184e-05],
            [+3.9008038585852290e-05, -8.6543337237836941e-05, +1.0086268708372256e-04],
            [-2.1792111331028049e-05, +2.8414264003578678e-05, +1.8916279332120321e-05],
            [-3.8847190487495055e-06, +6.4069296782535270e-06, -1.7362611928711208e-05],
            [-1.5358378150451566e-05, -5.7611513737809471e-06, -1.9871764856476196e-05],
            [-3.7225957329162548e-05, -1.5687552833166010e-06, -3.4717349142275990e-05],
            [+3.2408291983641028e-07, -6.7683941908276048e-05, -6.1602680430568183e-05],
            [-8.1819149671629329e-06, -9.5470867672313039e-06, -1.7007294079721094e-05],
            [-1.7245516321582448e-05, +1.6416848978506148e-05, +1.6005589814916458e-05],
            [-3.0116095523698349e-06, +7.3382310356380993e-07, -1.3652036487357768e-05],
            [+1.0378663330288130e-04, +4.4180555120489236e-05, +3.2576724474326750e-05],
            [-2.8021453980080289e-05, +3.9008636389849655e-05, -1.2083469461741150e-04],
            [-3.0452136766248191e-05, +7.0712047105794630e-05, +8.3682948356528971e-05],
            [+1.6902554427499383e-06, -2.9147110970219634e-05, +3.6402009185785855e-05],
            [+4.6966237842630462e-05, -1.2554523547742754e-05, +3.0600183501647347e-05],
            [-1.6417061357004406e-05, +1.9834934423075429e-05, +1.3993399653561848e-05],
        ]
    )

    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-03", return_dict=True)),
            specification={
                "keywords": {"level_hint": "D4"},
                "driver": "gradient",
                "model": {"method": "r2scan"},
            },
        )
    else:
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("mindless-03", return_dict=True)),
            keywords={"level_hint": "D4"},
            driver="gradient",
            model={"method": "r2scan"},
        )

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    atomic_result = qcng.compute(atomic_input, "dftd4", return_version=retver)
    atomic_result = checkver_and_convert(atomic_result, request.node.name, "post")

    print(atomic_result.return_result)
    assert atomic_result.success
    assert pytest.approx(atomic_result.return_result, abs=thr) == return_result


@uusing("dftd4")
def test_dftd4_task_unknown_method(schema_versions, request):
    models, retver, models_out = schema_versions

    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            specification={
                "keywords": {"level_hint": "D4"},
                "model": {"method": "non-existent-method"},
                "driver": "energy",
            },
        )
    else:
        atomic_input = models.AtomicInput(
            molecule=models.Molecule(**qcng.get_molecule("water", return_dict=True)),
            keywords={"level_hint": "D4"},
            model={"method": "non-existent-method"},
            driver="energy",
        )
    error = models_out.ComputeError(
        error_type="input error", error_message="Functional 'non-existent-method' not known"
    )

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    atomic_result = qcng.compute(atomic_input, "dftd4", return_version=retver)
    atomic_result = checkver_and_convert(atomic_result, request.node.name, "post", vercheck=False)

    print(atomic_result.error)
    assert not atomic_result.success
    assert atomic_result.error == error


@uusing("dftd4")
def test_dftd4_task_cold_fusion(schema_versions, request):
    models, retver, models_out = schema_versions

    li4 = {
        "symbols": ["Li", "Li", "Li", "Li"],
        "geometry": [
            [-1.58746019997201, +1.58746019997201, +1.58746019997201],
            [-1.58746019997201, +1.58746019997201, +1.58746019997201],
            [-1.58746019997201, -1.58746019997201, -1.58746019997201],
            [+1.58746019997201, +1.58746019997201, -1.58746019997201],
        ],
        "validated": True,  # Force a nuclear fusion input, to make dftd4 fail
    }

    if from_v2(request.node.name):
        atomic_input = models.AtomicInput(
            molecule=li4,
            specification={
                "keywords": {"level_hint": "D4"},
                "model": {"method": "pbe"},
                "driver": "energy",
            },
        )
    else:
        atomic_input = models.AtomicInput(
            molecule=li4,
            keywords={"level_hint": "D4"},
            model={"method": "pbe"},
            driver="energy",
        )
    error = models_out.ComputeError(
        error_type="input error",
        error_message="Too close interatomic distances found",
    )

    atomic_input = checkver_and_convert(atomic_input, request.node.name, "pre")
    atomic_result = qcng.compute(atomic_input, "dftd4", return_version=retver)
    atomic_result = checkver_and_convert(atomic_result, request.node.name, "post", vercheck=False)

    print(atomic_result.error)
    assert not atomic_result.success
    assert atomic_result.error == error



"""
Tests the DQM compute dispatch module
"""
import pprint
import re
import sys
import tempfile
from pathlib import Path

import pytest

import qcengine as qcng
from qcengine.testing import checkver_and_convert, from_v2, schema_versions, using

_canonical_methods = [
    # needs attn ("adcc", {"method": "adc2", "basis": "6-31G"}, {"n_triplets": 3}),
    pytest.param("cfour", {"method": "hf", "basis": "6-31G"}, {}, marks=using("cfour")),
    pytest.param("dftd3", {"method": "b3lyp-d3"}, {}, marks=using("classic-dftd3")),
    pytest.param("gamess", {"method": "hf", "basis": "n31"}, {"basis__NGAUSS": 6}, marks=using("gamess")),
    pytest.param("jaguar", {"method": "hf", "basis": "6-31G"}, {"nops": 1}, marks=using("jaguar")),
    pytest.param("gcp", {"method": "hf3c"}, {}, marks=using("classic-gcp")),
    pytest.param("mctc-gcp", {"method": "dft/sv"}, {}, marks=using("mctc-gcp")),
    # needs attn ("molpro", {"method": "hf", "basis": "6-31G"}, {}),
    # needs attn ("mopac", {"method": "PM6"}, {}),
    pytest.param("mp2d", {"method": "MP2-DMP2"}, {}, marks=using("mp2d")),
    # needs attn ("mrchem", {"method": "blyp"}, {"world_prec": 1.0e-3}),
    pytest.param("nwchem", {"method": "hf", "basis": "6-31G"}, {}, marks=using("nwchem")),
    # needs attn ("openmm", {"method": "openff-1.0.0", "basis": "smirnoff"}, {}),
    pytest.param("psi4", {"method": "hf", "basis": "6-31G"}, {"gradient_write": True}, marks=using("psi4")),
    # needs attn ("qchem", {"method": "hf", "basis": "6-31G"}, {}),
    # needs attn ("qcore", {"method": "pbe", "basis": "6-31G"}, {}),
    # needs attn ("rdkit", {"method": "UFF"}, {}),
    # needs attn ("terachem", {"method": "bad"}, {}),
    # needs attn ("terachem_pbs", {"method": "b3lyp", "basis": "6-31G"}, {}),
    # needs attn ("torchani", {"method": "ANI1x"}, {}),
    # needs attn ("turbomole", {"method": "pbe", "basis": "6-31G"}, {}),
    # needs attn ("xtb", {"method": "GFN2-xTB"}, {}),
]


def _get_molecule(program, method, molcls):
    if program in ["openmm", "terachem_pbs"]:
        dmol = qcng.get_molecule("water", return_dict=True)
    elif program == "gamess" and method == "ccsd(t)":
        dmol = qcng.get_molecule("water", return_dict=True)
    else:
        dmol = qcng.get_molecule("hydrogen", return_dict=True)

    return molcls(**dmol)


@pytest.mark.parametrize(
    "memory_trickery",
    [
        pytest.param({}, id="none"),
        pytest.param(
            # native keywords that CONTRADICT config.memory below
            {
                "cfour": {"memory_size": "5000"},
                "gamess": {"system__mwords": "5000"},
                "nwchem": {"memory": "5 gb"},
                "psi4": {},  # no contradictory memory keyword in psi
            },
            id="dsl",
        ),
    ],
)
@pytest.mark.parametrize("program, model, keywords", _canonical_methods)
def test_local_options_memory_gib(program, model, keywords, memory_trickery, schema_versions, request):
    """Ensure memory handling implemented in harness (if applicable).

    For available harnesses, run minimal calc at specific total node memory, both through runtime
      config alone and with clashing (and non-QCEngine-like) keyword spec. Check memory quantity
      shows up in ``TaskConfig``.
    For ``managed-memory``-active harnesses, check that memory registers in output.

    New Harness Instructions
    ------------------------
    * Make sure minimal calc is in _canonical_methods above.
    * If ``managed_memory=True`` in harness, add regex to ``stdout_ref`` below to check that memory
      is specifiable.
    * If this test doesn't work, implement or adjust ``config.memory`` in your harness.

    """
    models, retver, _ = schema_versions

    harness = qcng.get_program(program)
    molecule = _get_molecule(program, model["method"], models.Molecule)

    addl_keywords = memory_trickery.get(program, memory_trickery)
    use_keywords = {**keywords, **addl_keywords}

    #  <<  Config

    config = qcng.config.get_config(
        hostname="something",
        task_config={
            "ncores": 1,
            "nnodes": 1,
            "memory": 1.555,
        },
    )

    #  <<  Run

    if from_v2(request.node.name):
        inp = models.AtomicInput(
            molecule=molecule,
            specification=models.AtomicSpecification(driver="energy", model=model, keywords=use_keywords),
        )
    else:
        inp = models.AtomicInput(molecule=molecule, driver="energy", model=model, keywords=use_keywords)

    inp = checkver_and_convert(inp, request.node.name, "pre")
    ret = qcng.compute(inp, program, raise_error=True, task_config=config.model_dump(), return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    pprint.pprint(ret.model_dump(), width=200)
    assert ret.success is True

    #  <<  Reference

    stdout_ref = {  # 1.555 GiB = 208708567 quad-words
        "cfour": "Allocated    1592 MB of main memory",
        "gamess": "208000000 WORDS OF MEMORY AVAILABLE",
        "nwchem": r"total\s+=\s+2087085\d\d doubles =\s+1592.3 Mbytes",  # doubles is quad-words. Mbytes is MiB
        "psi4": "1592 MiB Core",
    }

    #  <<  Test

    assert config.ncores == 1
    assert pytest.approx(config.memory, 0.1) == 1.555

    if harness._defaults["managed_memory"] is True:
        assert re.search(stdout_ref[program], ret.stdout), f"Memory pattern not found: {stdout_ref[program]}"


@pytest.mark.parametrize("program, model, keywords", _canonical_methods)
def test_local_options_scratch(program, model, keywords, schema_versions, request):
    """Ensure scratch handling implemented in harness (if applicable).

    For available harnesses, run minimal calc at specific scratch directory name (randomly generated
      during test) and skip scratch clean-up. Check scratch settings show up in ``TaskConfig``.
    For ``scratch``-active harnesses, check that an expected file is written to and left behind in
      scratch directory. Check any scratch-related printing in output.

    New Harness Instructions
    ------------------------
    * Make sure minimal calc is in _canonical_methods above.
    * If ``scratch=True`` in harness, add single file (preferrably output) glob to ``scratch_sample``
      below to check that program scratch is directable.
    * If ``scratch=True`` in harness, if scratch directory mentioned in output, add regex to
      ``stdout_ref`` below to check that program scratch is directable. Otherwise, add an
      always-passing regex.
    * If this test doesn't work, implement or adjust ``config.scratch_directory`` and
      ``config.scratch_messy`` in your harness.

    """
    models, retver, _ = schema_versions

    harness = qcng.get_program(program)
    molecule = _get_molecule(program, model["method"], models.Molecule)

    #  <<  Config

    scratch_directory = tempfile.mkdtemp(suffix="_" + program)

    config = qcng.config.get_config(
        hostname="something",
        task_config={
            "scratch_directory": scratch_directory,
            "scratch_messy": True,
        },
    )

    #  <<  Run

    if from_v2(request.node.name):
        inp = models.AtomicInput(
            molecule=molecule, specification=models.AtomicSpecification(driver="energy", model=model, keywords=keywords)
        )
    else:
        inp = models.AtomicInput(molecule=molecule, driver="energy", model=model, keywords=keywords)

    inp = checkver_and_convert(inp, request.node.name, "pre")
    ret = qcng.compute(inp, program, raise_error=True, task_config=config.model_dump(), return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    pprint.pprint(ret.model_dump(), width=200)
    assert ret.success is True

    #  <<  Reference

    stdout_ref = {
        "cfour": "University of Florida",  # freebie
        "dftd3": "Grimme",  # freebie
        "gamess": "IOWA STATE UNIVERSITY",  # freebie
        "jaguar": "Jaguar version",  # freebie
        "gcp": "Grimme",  # freebie
        "mctc-gcp": "Grimme",  # freebie
        "mp2d": "Beran",  # freebie
        "nwchem": "E. Apra",  # freebie
        "psi4": rf"Scratch directory: {scratch_directory}/tmp\w+_psi_scratch/",
    }
    if sys.platform.startswith("win"):
        # too hard to regex Windows paths that need escape chars
        stdout_ref["psi4"] = f"Scratch directory: "

    # a scratch file (preferrably output) expected after job if scratch not cleaned up
    scratch_sample = {
        "cfour": "*/NEWFOCK",
        "dftd3": "*/dftd3_geometry.xyz",  # no outfiles
        "gamess": "*/gamess.dat",
        "jaguar": "*/dispatch.out",
        "gcp": "*/gcp_geometry.xyz",  # no outfiles
        "mctc-gcp": "*/gcp_geometry.xyz",  # no outfiles
        "mp2d": "*/mp2d_geometry",  # no outfiles
        "nwchem": "*/nwchem.db",
        "psi4": "*/psi.*.35",
    }

    #  <<  Test

    assert config.scratch_directory.endswith(program)

    if harness._defaults["scratch"] is True:
        sample_file = list(Path(scratch_directory).glob(scratch_sample[program]))
        assert len(sample_file) == 1, f"Scratch sample not found: {scratch_sample[program]} in {scratch_directory}"

        assert re.search(stdout_ref[program], ret.stdout), f"Scratch pattern not found: {stdout_ref[program]}"


@pytest.mark.parametrize("ncores", [1, 3])
@pytest.mark.parametrize("program, model, keywords", _canonical_methods)
def test_local_options_ncores(program, model, keywords, ncores, schema_versions, request):
    """Ensure multithreading implemented in harness (if applicable) or multithreaded runs don't
    break harness (if inapplicable).

    For available harnesses, run minimal calc with single and multiple cores; check ncores count
      shows up in ``TaskConfig``.
    For ``thread_parallel``-active harnesses, check ncores count registers in output.

    New Harness Instructions
    ------------------------
    * Make sure minimal calc is in _canonical_methods above.
    * If ``thread_parallel=True`` in harness, add regex to ``stdout_ref`` below to check ncores the
      program sees.
    * If this test doesn't work, implement or adjust ``config.ncores`` in your harness.

    """
    models, retver, _ = schema_versions

    harness = qcng.get_program(program)
    molecule = _get_molecule(program, model["method"], models.Molecule)

    #  <<  Config

    config = qcng.config.get_config(
        hostname="something",
        task_config={
            "ncores": ncores,
            "nnodes": 1,
        },
    )

    #  <<  Run

    if from_v2(request.node.name):
        inp = models.AtomicInput(
            molecule=molecule, specification=models.AtomicSpecification(driver="energy", model=model, keywords=keywords)
        )
    else:
        inp = models.AtomicInput(molecule=molecule, driver="energy", model=model, keywords=keywords)

    inp = checkver_and_convert(inp, request.node.name, "pre")
    ret = qcng.compute(inp, program, raise_error=True, task_config=config.model_dump(), return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    pprint.pprint(ret.model_dump(), width=200)
    assert ret.success is True

    #  <<  Reference

    stdout_ref = {
        "cfour": rf"Running with {ncores} threads/proc",
        "gamess": rf"MEMDDI DISTRIBUTED OVER\s+{ncores} PROCESSORS",
        "jaguar": rf"Using up to\s+{ncores} threads per process",
        # "gamess": rf"PARALLEL VERSION RUNNING ON\s+{ncores} PROCESSORS IN\s+1 NODES",  # no line for serial
        # nwchem is node_parallel only
        "psi4": rf"Threads:\s+{ncores}",
    }

    #  <<  Test

    assert config.ncores == ncores
    assert config.nnodes == 1

    if harness._defaults["thread_parallel"] is True:
        assert re.search(stdout_ref[program], ret.stdout), f"Thread pattern not found: {stdout_ref[program]}"



import abc
import logging
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, ConfigDict

from qcengine.config import TaskConfig
from qcengine.exceptions import InputError, KnownErrorException

from ..util import model_wrapper

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from qcelemental.models.v2 import AtomicInput, AtomicResult, FailedOperation


class ProgramHarness(BaseModel, abc.ABC):
    """Base class for analytic single-geometry capable harnesses."""

    _defaults: ClassVar[Dict[str, Any]] = {}
    name: str
    scratch: bool
    thread_safe: bool
    thread_parallel: bool
    node_parallel: bool
    managed_memory: bool
    extras: Optional[Dict[str, Any]] = None

    model_config = ConfigDict(
        frozen=True,
        extra="forbid",
    )

    def __init__(self, **kwargs):
        super().__init__(**{**self._defaults, **kwargs})

    @abc.abstractmethod
    def compute(self, input_data: "AtomicInput", config: TaskConfig) -> Union["AtomicResult", "FailedOperation"]:
        """Top-level compute method to be implemented for every ProgramHarness

        Note:
            This method behave in any of the following ways:
                1. Return AtomicResult upon successful completion of a calculation
                2. Return FailedOperation object if an operation was unsuccessful or raised an exception. This is most
                    likely to occur if the underlying QC package has a QCSchema API that catches exceptions and
                    returns them as FailedOperation objects to end users.
                3. Raise an exception if a computation failed. The raised exception will be handled by the
                    qcng.compute() method and either raised or packaged as a FailedOperation object.
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def found(raise_error: bool = False) -> bool:
        """
        Checks if the program can be found.

        Parameters
        ----------
        raise_error : bool, optional
            If True, raises an error if the program cannot be found.

        Returns
        -------
        bool
            Returns True if the program was found, False otherwise.
        """

    ## Utility

    def build_input_model(
        self, data: Dict[str, Any], *, return_input_schema_version: bool = False
    ) -> Union["AtomicInput", Tuple["AtomicInput", int]]:
        """
        Quick wrapper around util.model_wrapper for inherited classes
        """
        # Note: Someday when the multiple QCSchema versions QCEngine supports are all within the
        #  Pydantic v2 API base class, this can use discriminated unions instead of logic.

        from qcelemental.models.v1 import AtomicInput as v1_model
        from qcelemental.models.v2 import AtomicInput as v2_model

        if isinstance(data, v1_model):
            mdl = model_wrapper(data, v1_model)
        elif isinstance(data, v2_model):
            mdl = model_wrapper(data, v2_model)
        elif isinstance(data, dict):
            # remember these are user-provided dictionaries, so they'll have the mandatory fields,
            #   like driver, not the helpful discriminator fields like schema_version.
            # when dictionaries look the same, we can't correctly identify the user schema version
            #   so have to default to one or the other. Note that can force paths for testing by
            #   -1 -> 2 in schema_versions.
            # so long as versions distinguishable by a *required* field, id by dict is reliable.

            if data.get("specification", False) or data.get("schema_version") == 2:
                mdl = model_wrapper(data, v2_model)
            else:
                try:
                    mdl = model_wrapper(data, v1_model)
                except RuntimeError:
                    # form a QCSchema v1 layout model with Pydantic v2 API
                    # this is for py314 and only safe b/c immediately converted to v2 next
                    from qcelemental.models._v1v2 import AtomicInput as v1v2_model

                    mdl = model_wrapper(data, v1v2_model)
        else:
            raise InputError(f"Unsupported input type for build_input_model: {type(data).__name__}")

        input_schema_version = mdl.schema_version
        if return_input_schema_version:
            return mdl.convert_v(2), input_schema_version
        else:
            return mdl.convert_v(2)

    def get_version(self) -> str:
        """Finds program, extracts version, returns normalized version string.

        Returns
        -------
        str
            Return a valid, safe python version string.
        """

    ## Computers

    def build_input(
        self, input_model: "AtomicInput", config: TaskConfig, template: Optional[str] = None
    ) -> Dict[str, Any]:
        raise ValueError("build_input is not implemented for {}.", self.__class__)

    def execute(
        self,
        inputs: Dict[str, Any],
        extra_outfiles: Optional[List[str]] = None,
        extra_commands: Optional[List[str]] = None,
        scratch_name: Optional[str] = None,
        timeout: Optional[int] = None,
    ) -> Tuple[bool, Dict[str, Any]]:
        raise ValueError("execute is not implemented for {}.", self.__class__)

    def parse_output(self, outfiles: Dict[str, str], input_model: "AtomicInput") -> "AtomicResult":
        raise ValueError("parse_output is not implemented for {}.", self.__class__)


class ErrorCorrectionProgramHarness(ProgramHarness, abc.ABC):
    """Base class for Harnesses that include logic to correct common errors

    Classes which implement this Harness must override the :meth:`_compute` method
    rather than :meth:`compute`. The ``compute`` method from this class will make
    calls to ``_compute`` with different actions as it attempts to correct errors in the
    input files.

    The error corrections are defined by first implementing a :class:`KnownErrorException`
    that contains logic to determine if a certain error occurs and returns an appropriate
    update to the keywords of an AtomicInput.

    Then, modify the ``_compute`` method of your Harness to run the ``detect_error`` method
    of each ``KnownErrorException`` for the Harness after it finishes performing the computation.
    The ``detect_error`` method will raise an exception that is caught by the
    ``ErrorCorrectionProgramHarness`` and used to determine if/how to re-run the computation.
    """

    def _compute(self, input_data: "AtomicInput", config: TaskConfig) -> "AtomicResult":
        raise NotImplementedError()

    def compute(self, input_data: "AtomicInput", config: TaskConfig) -> "AtomicResult":
        # Get the error correction configuration
        error_policy = input_data.specification.protocols.error_correction

        # Create a local copy of the input data
        local_input_data = input_data

        # Run the method and, if it fails, assess if the failure is restartable
        observed_errors = {}  # Errors that have been observed previously
        while True:
            try:
                result = self._compute(local_input_data, config)
                break
            except KnownErrorException as e:
                logger.info(f"Caught a {type(e)} error.")

                # Determine whether this specific type of error is allowed
                correction_allowed = error_policy.allows(e.error_name)
                if not correction_allowed:
                    logger.info(f'Error correction for "{e.error_name}" is not allowed')
                    raise e
                logger.info(f'Error correction for "{e.error_name}" is allowed')

                # Check if it has run before
                # TODO (wardlt): Should we allow errors to be run >1 time?
                previously_run = e.error_name in observed_errors
                if previously_run:
                    logger.info(
                        "Error has been observed before and mitigation did not fix the issue. Raising exception"
                    )
                    raise e

                # Generate and apply the updated keywords
                keyword_updates = e.create_keyword_update(local_input_data)
                new_keywords = local_input_data.specification.keywords.copy()
                new_keywords.update(keyword_updates)
                local_input_data = input_data.__class__(
                    **local_input_data.model_dump(exclude={"specification"}),
                    specification={**local_input_data.specification.model_dump(), "keywords": new_keywords},
                )

                # Store the error details and mitigations employed
                observed_errors[e.error_name] = {"details": e.details, "keyword_updates": keyword_updates}

        # Add the errors observed and corrected for, if any
        if len(observed_errors) > 0:
            result.extras["observed_errors"] = observed_errors
        return result



import pprint
import re

import pytest
import qcelemental as qcel

import qcengine as qcng
from qcengine.testing import checkver_and_convert, from_v2, schema_versions

from .test_canonical_config import _canonical_methods, _get_molecule


@pytest.mark.parametrize("native", ["none", "input", "all"])
@pytest.mark.parametrize("program, model, keywords", _canonical_methods)
def test_protocol_native(program, model, keywords, native, schema_versions, request):
    """Ensure native_files protocol implemented in harness.

    For harnesses, run minimal gradient calc with different protocol settings; check expected
      native/DSL files show up in ``AtomicResult``.

    New Harness Instructions
    ------------------------
    * Make sure minimal calc is in _canonical_methods above. This uses gradient for more files.
    * Add regex to ``input_ref`` & ``all_ref`` below to check content of input and any other file.
    * If this test doesn't work, implement or adjust ``native_files`` in your harness.

    """
    models, retver, _ = schema_versions

    harness = qcng.get_program(program)
    molecule = _get_molecule(program, model["method"], models.Molecule)

    #  <<  Config

    protocols = {
        "native_files": native,
    }
    config = qcng.config.get_config(
        hostname="something",
        task_config={
            "ncores": 1,
            "nnodes": 1,
        },
    )

    #  <<  Run

    if from_v2(request.node.name):
        inp = models.AtomicInput(
            molecule=molecule,
            specification=models.AtomicSpecification(
                driver="gradient", model=model, keywords=keywords, protocols=protocols
            ),
        )
    else:
        inp = models.AtomicInput(
            molecule=molecule, driver="gradient", model=model, keywords=keywords, protocols=protocols
        )

    inp = checkver_and_convert(inp, request.node.name, "pre")
    ret = qcng.compute(inp, program, raise_error=True, task_config=config.model_dump(), return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    pprint.pprint(ret.model_dump(), width=200)
    assert ret.success is True

    #  <<  Reference

    input_ref = {
        "cfour": rf"\*CFOUR\(BASIS=6-31G",
        "dftd3": rf"1.000000     1.261000     1.703000",
        "gamess": rf"\$basis gbasis=n31 ngauss=6 \$end",
        "jaguar": rf"basis=6-31G",
        "gcp": rf"level HF3C",
        "mctc-gcp": rf"level DFT/SV",
        "mp2d": rf"--TT_a1=0.944 --TT_a2=0.48",
        "nwchem": rf"H library 6-31G",
        "psi4": rf'"driver": "gradient",',
    }

    all_ref = {
        "cfour": ("GRD", rf"1.0+\s+0.0+\s+0.0+\s+0.03"),
        "dftd3": ("dftd3_geometry.xyz", rf"H\s+0.0+\s+0.0+\s+0.34"),
        "gamess": ("gamess.dat", rf"CLOSED SHELL ORBITALS --- GENERATED AT"),
        "jaguar": ("dispatch.01.in", rf"&gen"),
        "gcp": ("gcp_geometry.xyz", rf"H\s+0.0+\s+0.0+\s+0.34"),
        "mctc-gcp": ("gcp_geometry.xyz", rf"H\s+0.0+\s+0.0+\s+0.34"),
        "mp2d": ("mp2d_geometry", rf"H\s+0.0+\s+0.0+\s+0.34"),
        "nwchem": ("nwchem.grad", rf"0.0, 0.0, 0.03"),
        "psi4": ("psi4.grad", rf"1.0+\s+(-?)0.0+\s+(-?)0.0+\s+0.03"),
    }

    #  <<  Test

    if native == "none":
        if qcel.util.parse_version(qcel.__version__) < qcel.util.parse_version("0.25.0"):
            assert ret.native_files is None
        else:
            assert ret.native_files == {}
    elif native == "input":
        assert list(ret.native_files.keys()) == ["input"]

    if ret.native_files:
        assert "stdout" not in ret.native_files, f"Stdout found in native_files -- clean up the harness"
        assert "stderr" not in ret.native_files, f"Stderr found in native_files -- clean up the harness"
    assert "outfiles" not in ret.extras, f"Outfiles found in extras -- clean up the harness"

    if native in ["input", "all"]:
        assert re.search(
            input_ref[program], ret.native_files["input"]
        ), f"Input file pattern not found: {input_ref[program]}"
    if native == "all" and program != "psi4":  # allow psi4 once native_files PR merged
        fl, snip = all_ref[program]
        assert re.search(snip, ret.native_files[fl]), f"Ancillary file pattern not found in {fl}: {snip}"



from typing import Any, ClassVar, Dict, Union

from qcelemental.models._v1v2 import OptimizationResult
from qcelemental.models.v2 import OptimizationInput
from qcelemental.util import safe_version, which_import

from ..exceptions import RandomError, UnknownError
from ..util import QCEL_V1V2_SHIM_CODE, environ_context
from .model import ProcedureHarness


def _handle_errors(output_data: Dict[str, Any], default_error: str) -> tuple[str, str]:
    error_block = output_data.get("error")
    if isinstance(error_block, dict):
        if "error_message" in error_block:
            return error_block["error_message"], error_block.get("error_type", "unknown_error")
        return str(error_block), "unknown_error"
    if isinstance(error_block, str):
        return error_block, "unknown_error"
    return default_error, "unknown_error"


def _attach_retry_provenance(failed_result: Dict[str, Any], retries: int) -> None:
    trajectory = failed_result.get("trajectory")
    if isinstance(trajectory, list):
        for point in trajectory:
            if isinstance(point, dict):
                point.setdefault("provenance", {}).setdefault("retries", retries)


class GeometricProcedure(ProcedureHarness):

    _defaults: ClassVar[Dict[str, Any]] = {"name": "geomeTRIC", "procedure": "optimization"}

    version_cache: Dict[str, str] = {}

    def found(self, raise_error: bool = False) -> bool:
        return which_import(
            "geometric",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via `conda install geometric -c conda-forge`.",
        )

    def get_version(self) -> str:
        self.found(raise_error=True)

        which_prog = which_import("geometric")
        if which_prog not in self.version_cache:
            import geometric

            self.version_cache[which_prog] = safe_version(geometric.__version__)

        return self.version_cache[which_prog]

    def build_input_model(
        self, data: Union[Dict[str, Any], "OptimizationInput"], *, return_input_schema_version: bool = False
    ) -> "OptimizationInput":
        return self._build_model(data, "OptimizationInput", return_input_schema_version=return_input_schema_version)

    def compute(self, input_model: "OptimizationInput", config: "TaskConfig") -> "OptimizationResult":
        try:
            import geometric
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Could not find geomeTRIC in the Python path.")

        input_data_v2 = input_model.model_dump()

        # Temporary patch for geomeTRIC
        input_data_v2["initial_molecule"]["symbols"] = list(input_data_v2["initial_molecule"]["symbols"])

        # Set retries to two if zero while respecting local_config
        local_config = config.model_dump()
        local_config["retries"] = local_config.get("retries", 2) or 2
        input_data_v2["specification"]["specification"]["extras"]["_qcengine_local_config"] = local_config

        # Run the program
        # * geomeTRIC only speaks v1 as of v1.1
        # * envvar allows geomeTRIC to call back QCEngine for gradients
        input_data__v1v2 = OptimizationInput(**input_data_v2).convert_v(QCEL_V1V2_SHIM_CODE).model_dump()
        with environ_context(env={"QCNG_USE_V1V2_SHIM": "1"}):
            output_v1 = geometric.run_json.geometric_run_json(input_data__v1v2)

        output_v1["provenance"] = {
            "creator": "geomeTRIC",
            "routine": "geometric.run_json.geometric_run_json",
            "version": geometric.__version__,
        }

        output_v1["schema_name"] = "qcschema_optimization_output"  # overwrites OptIn value
        output_v1["input_specification"]["extras"].pop("_qcengine_local_config", None)
        if output_v1["success"]:
            output_model__v1v2 = OptimizationResult(**output_v1)
            output = output_model__v1v2.convert_v(2, external_input_data=input_model)
        else:
            error_message, error_type = _handle_errors(output_v1, "geomeTRIC failed without an error message.")
            is_random_path = error_type == "random_error"
            if is_random_path:
                _attach_retry_provenance(output_v1, local_config["retries"])
            error_cls = RandomError if is_random_path else UnknownError
            raise error_cls(error_message, extras={"failed_result": output_v1})

        return output



"""
Provides a CLI for QCEngine
"""

import argparse
import json
import os.path
import sys
from typing import Any, Dict

from . import get_program  # run and run-procedure; info
from . import (
    __version__,
    compute,
    compute_procedure,
    get_procedure,
    list_all_procedures,
    list_all_programs,
    list_available_procedures,
    list_available_programs,
)
from .config import global_repr  # info

__all__ = ["main"]

info_choices = frozenset(["version", "programs", "procedures", "config", "bulletin", "all"])


def parse_args():
    parser = argparse.ArgumentParser(description="A CLI for the QCEngine.")
    parser.add_argument("--version", action="version", version=f"{__version__}")

    parent_parser = argparse.ArgumentParser(add_help=False)
    task_group = parent_parser.add_argument_group(
        "Task Configuration", "Extra configuration related to running the computation"
    )
    task_group.add_argument("--ncores", type=int, help="The number of cores to use for the task")
    task_group.add_argument("--nnodes", type=int, help="The number of nodes to use")
    task_group.add_argument("--memory", type=float, help="The amount of memory (in GiB) to use")
    task_group.add_argument("--scratch-directory", type=str, help="Where to store temporary files")
    task_group.add_argument("--retries", type=int, help="Number of retries for random failures")
    task_group.add_argument("--mpiexec-command", type=str, help="Command used to launch MPI tasks")
    task_group.add_argument(
        "--use-mpiexec",
        action="store_true",
        default=None,
        help="Whether it is necessary to use MPI to run an executable",
    )
    task_group.add_argument("--cores-per-rank", type=int, help="Number of cores per MPI rank")
    task_group.add_argument(
        "--scratch-messy",
        action="store_true",
        default=None,
        help="Leave the scratch directory and contents on disk after completion",
    )

    subparsers = parser.add_subparsers(dest="command")

    info = subparsers.add_parser("info", help="Print information about QCEngine setup, version, and environment.")
    info.add_argument(
        "category", nargs="*", default="all", choices=info_choices, help="The information categories to show."
    )

    run = subparsers.add_parser(
        "run", parents=[parent_parser], help="Run a program on a given task. Output is printed as a JSON blob."
    )
    run.add_argument("program", type=str, help="The program to run.")
    run.add_argument(
        "data",
        type=str,
        help="Data describing the task to run. "
        "One of: (i) A JSON blob, "
        "(ii) A file name, "
        "(iii) '-', indicating data will be read from STDIN.",
    )
    run.add_argument(
        "--return-version", default=-1, type=int, help="Schema version to return, if not the input version"
    )

    run_procedure = subparsers.add_parser(
        "run-procedure",
        parents=[parent_parser],
        help="Run a procedure on a given task. Output is printed as a JSON blob. "
        "Since v0.50.0, favor 'run', even for procedures.",
    )
    run_procedure.add_argument("procedure", type=str, help="The procedure to run.")
    run_procedure.add_argument(
        "data",
        type=str,
        help="Data describing the task to run. "
        "One of: (i) A JSON blob, "
        "(ii) A file name, "
        "(iii) '-', indicating data will be read from STDIN.",
    )

    args = vars(parser.parse_args())
    if args["command"] is None:
        parser.print_help(sys.stderr)
        exit(1)

    return args


def info_cli(args):
    def info_version():
        import qcelemental

        print(">>> Version information")
        print(f"QCEngine:    {__version__}")
        print(f"QCElemental: {qcelemental.__version__}")
        print()

    def info_programs():  # lgtm: [py/similar-function]
        print(">>> Program information")
        all_progs = list_all_programs()
        avail_progs = list_available_programs()
        print("Available programs:")
        for prog_name in sorted(avail_progs):
            program = get_program(prog_name)
            version = program.get_version()
            for loc, ver in program.version_cache.items():
                if ver == version:
                    which = loc
                    break
            else:
                which = "???"
            if version is None:
                version = "???"
            print(f"{prog_name + ':':12} v{version:20} {which}")

        print()
        print("Other supported programs:")
        print(" ".join(sorted(all_progs - avail_progs)))
        print()
        print(
            """If you think available programs are missing, query for details: `python -c "import qcengine as qcng; qcng.get_program('<program>')"`"""
        )
        print()

    def info_procedures():  # lgtm: [py/similar-function]
        print(">>> Procedure information")
        all_procs = list_all_procedures()
        avail_procs = list_available_procedures()
        print("Available procedures:")
        for proc_name in sorted(avail_procs):
            version = get_procedure(proc_name).get_version()
            if version is None:
                version = "???"
            print(f"{proc_name} v{version}")

        print()
        print("Other supported procedures:")
        print(" ".join(sorted(all_procs - avail_procs)))
        print()

    def info_bulletin():
        print(">>> Bulletin")
        bull_items = []
        bull_items.append(
            """  * [Apr 2025; v0.32.0] Program Harness 'qcore'/'entos' is deprecated with Prof. Manby's approval.
     It will cease to be tested as soon as Python minimum bumps >3.7. It may be removed as soon as 2026."""
        )
        bull_items.append(
            """  * [Apr 2025; v0.32.0] Program Harness 'dftd3' is deprecated in favor of the maintained 's-dftd3'.
     Its testing will become narrower (w/o Psi4). It may be removed as soon as 2026."""
        )
        bull_items.append(
            """  * [Apr 2025; v0.32.0] Program Harness 'gcp' is deprecated in favor of the maintained 'mctc-gcp'.
     Its testing will become narrower (w/o Psi4). It may be removed as soon as 2026."""
        )
        print("\n".join(bull_items))
        print()

    # default=["all"] does is not allowed by argparse
    if not isinstance(args["category"], list):
        args["category"] = [args["category"]]
    cat = set(args["category"])

    if "version" in cat or "all" in cat:
        info_version()
    if "programs" in cat or "all" in cat:
        info_programs()
    if "procedures" in cat or "all" in cat:
        info_procedures()
    if "bulletin" in cat or "all" in cat:
        info_bulletin()
    if "config" in cat or "all" in cat:
        print(">>> Configuration information")
        print()
        print(global_repr())


def data_arg_helper(data_arg: str) -> Dict[str, Any]:
    """
    Converts the data argument of run and run-procedure commands to a dict for compute or compute_procedure

    Parameters
    ----------
    data_arg: str
        Either a data blob or file name or '-' for STDIN

    Returns
    -------
    Dict[str, Any]
        An input for compute or compute_procedure.
    """
    if data_arg == "-":
        return json.load(sys.stdin)
    elif os.path.isfile(data_arg):
        return json.load(open(data_arg))
    else:
        return json.loads(data_arg)


def main(args=None):
    # Grab CLI args if not present
    if args is None:
        args = parse_args()

    # Break out a task config
    task_config = {
        "ncores": args.pop("ncores", None),
        "memory": args.pop("memory", None),
        "nnodes": args.pop("nnodes", None),
        "scratch_directory": args.pop("scratch_directory", None),
        "retries": args.pop("retries", None),
        "mpiexec_command": args.pop("mpiexec_command", None),
        "use_mpiexec": args.pop("use_mpiexec", None),
        "cores_per_rank": args.pop("cores_per_rank", None),
        "scratch_messy": args.pop("scratch_messy", None),
    }

    # Prune None values and let other config functions handle defaults
    task_config = {k: v for k, v in task_config.items() if v is not None}

    command = args.pop("command")
    if command == "info":
        info_cli(args)
    elif command == "run":
        ret = compute(
            data_arg_helper(args["data"]),
            args["program"],
            task_config=task_config,
            return_version=args["return_version"],
        )
        print(ret.json())
    elif command == "run-procedure":
        ret = compute_procedure(data_arg_helper(args["data"]), args["procedure"], task_config=task_config)
        print(ret.json())



import abc
import importlib
from typing import Any, Dict, Tuple, Union

from pydantic import BaseModel, ConfigDict

from ..util import model_wrapper


class ProcedureHarness(BaseModel, abc.ABC):

    name: str
    procedure: str

    model_config = ConfigDict(
        frozen=True,
        extra="forbid",
    )

    def __init__(self, **kwargs):
        super().__init__(**{**self._defaults, **kwargs})

    @abc.abstractmethod
    def build_input_model(self, data: Union[Dict[str, Any], "BaseModel"], raise_error: bool = True) -> "BaseModel":
        """
        Build and validate the input model, passes if the data was a normal BaseModel input.

        Parameters
        ----------
        data : Union[Dict[str, Any], 'BaseModel']
            A data blob to construct the model from or the input model itself
        raise_error : bool, optional
            Raise an error or not if the operation failed.

        Returns
        -------
        BaseModel
            The input model for the procedure.
        """

    @abc.abstractmethod
    def compute(self, input_data: "BaseModel", config: "TaskConfig") -> "BaseModel":
        pass

    @abc.abstractmethod
    def found(self, raise_error: bool = False) -> bool:
        """
        Checks if the program can be found.

        Returns
        -------
        bool
            If the proceudre was found or not.
        """

    def _build_model(
        self, data: Dict[str, Any], model: "BaseModel", /, *, return_input_schema_version: bool = False
    ) -> Union["BaseModel", Tuple["BaseModel", int]]:
        """
        Quick wrapper around util.model_wrapper for inherited classes
        """

        module_v1 = importlib.import_module("qcelemental.models.v1")
        module_v2 = importlib.import_module("qcelemental.models.v2")
        v1_model = getattr(module_v1, model)
        v2_model = getattr(module_v2, model)

        if isinstance(data, v1_model):
            mdl = model_wrapper(data, v1_model)
        elif isinstance(data, v2_model):
            mdl = model_wrapper(data, v2_model)
        elif isinstance(data, dict):
            # remember these are user-provided dictionaries, so they'll have the mandatory fields,
            #   like driver, not the helpful discriminator fields like schema_version.
            # so long as versions distinguishable by a *required* field, id by dict is reliable.

            if data.get("specification", False) or data.get("schema_version") == 2:
                mdl = model_wrapper(data, v2_model)
            else:
                mdl = model_wrapper(data, v1_model)

        input_schema_version = mdl.schema_version
        if return_input_schema_version:
            return mdl.convert_v(2), input_schema_version
        else:
            return mdl.convert_v(2)

    def get_version(self) -> str:
        """Finds procedure, extracts version, returns normalized version string.
        Returns
        -------
        str
            Return a valid, safe python version string.
        """



"""
Calls the Molpro executable.
"""

import string
from typing import Any, ClassVar, Dict, List, Optional, Set, Tuple
from xml.etree import ElementTree as ET

from qcelemental.models.v2 import AtomicResult
from qcelemental.util import parse_version, safe_version, which

from ..exceptions import InputError, UnknownError
from ..util import execute
from .model import ProgramHarness


class MolproHarness(ProgramHarness):
    """Interface for Molpro project."""

    _defaults: ClassVar[Dict[str, Any]] = {
        "name": "Molpro",
        "scratch": True,
        "thread_safe": False,
        "thread_parallel": True,
        "node_parallel": False,
        "managed_memory": True,
    }
    version_cache: Dict[str, str] = {}

    # Set of implemented dft functionals in Molpro according to dfunc.registry (version 2019.2)
    # fmt: off
    _dft_functionals: Set[str] = {
        "B86MGC", "B86R", "B86", "B88C", "B88", "B95", "B97DF", "B97RDF", "BR", "BRUEG", "BW", "CS1", "CS2", "DIRAC",
        "ECERFPBE", "ECERF", "EXACT", "EXERFPBE", "EXERF", "G96", "HCTH120", "HCTH147", "HCTH93", "HJSWPBEX", "LTA",
        "LYP", "M052XC", "M052XX", "M05C", "M05X", "M062XC", "M062XX", "M06C", "M06HFC", "M06HFX", "M06LC", "M06LX",
        "M06X", "M12C", "MK00B", "MK00", "P86", "PBEC", "PBESOLC", "PBESOLX", "PBEXREV", "PBEX", "PW86", "PW91C",
        "PW91X", "PW92C", "STEST", "TFKE", "TH1", "TH2", "TH3", "TH4", "THGFCFO", "THGFCO", "THGFC", "THGFL", "TPSSC",
        "TPSSX", "VSXC", "VW", "VWN3", "VWN5", "XC-M05-2X", "XC-M05", "XC-M06-2X", "XC-M06-HF", "XC-M06-L", "XC-M06",
        "XC-M08-HX", "XC-M08-SO", "XC-M11-L", "XC-SOGGA11", "XC-SOGGA11-X", "XC-SOGGA", "FRMTST", "LHF", "TLHF",
        "LXBECKE", "ELP", "NULL", "YTEST", "TREF2", "TREF", "TTEST", "GLE", "GREEN", "SRB88", "SRLYP", "LB94", "EI",
        "SAOP", "USER", "INE", "ECERF2", "ECERFINTER", "ECERFLOCAL2", "ECERFLOCAL", "EXERFLOCAL", "FC", "FCFO", "FCO",
        "FL", "XC-M11", "PBEXANAL", "PBECANAL", "PBESOLCANAL", "PBESOLXANAL", "EXSRLDA", "EXSRLPBE", "ECSRLPBE",
        "ECSRLLPBE", "ECSQRTLPBE", "ECMUDIVLPBE", "EXERFPHS", "ECLERFMUPBE", "ECERFERFCPBE", "ECSQRTLDA", "REVPBEX",
        "B", "B-LYP", "BLYP", "B-P", "BP86", "B-VWN", "B3LYP", "B3LYP3", "B3LYP5", "B88X", "B97", "B97R", "BECKE",
        "BH-LYP", "CS", "D", "HFB", "HFS", "LDA", "LSDAC", "LSDC", "KYP88", "MM05", "MM05-2X", "MM06", "MM06-2X",
        "MM06-L", "MM06-HF", "PBE", "PBE0", "PBE0MOL", "PBEREV", "PW91", "S", "S-VWN", "SLATER", "VS99", "VWN",
        "VWN80", 'M05', "M05-2X", "M06", "M06-2X", "M06-L", "M06-HF", "M08-HX", "M08-SO", "M11-L", "TPSS", "TPSSH",
        "M12HFC", "HJSWPBE", "HJSWPBEH", "TCSWPBE", "PBESOL"
    }
    # fmt: on

    # Different open-shell scenarios:
    # - Restricted references
    #       RHF-RMP2 (like RHF-UMP2 using Molpro's naming convention for CC methods)
    #       RHF-UCCSD
    #       RHF-UCCSD(T)
    #       RHF-RCCSD
    #       RHF-RCCSD(T)
    # - Unrestricted references (Only supported up to UMP2, no CC support)
    #       UHF-UMP2
    # NOTE: Unrestricted SCF methods must be specified by using keyword reference
    _hf_methods: Set[str] = {"HF", "RHF"}
    _restricted_post_hf_methods: Set[str] = {"MP2", "CCSD", "CCSD(T)"}  # RMP2, RCCSD, RCCSD(T)}
    # TODO Add keyword to specify unrestricted for WF method
    # _unrestricted_post_hf_methods: Set[str] = {"UMP2", "UCCSD", "UCCSD(T)"}
    _post_hf_methods: Set[str] = {*_restricted_post_hf_methods}

    def found(self, raise_error: bool = False) -> bool:
        return which(
            "molpro", return_bool=True, raise_error=raise_error, raise_msg="Please install via https://www.molpro.net/"
        )

    # TODO Consider changing this to use molpro --version instead of performing a full execute
    def get_version(self) -> str:
        self.found(raise_error=True)

        name_space = {"molpro_uri": "http://www.molpro.net/schema/molpro-output"}
        which_prog = which("molpro")
        if which_prog not in self.version_cache:
            success, output = execute(
                [which_prog, "version.inp", "-d", ".", "-W", "."],
                infiles={"version.inp": ""},
                outfiles=["version.out", "version.xml"],
            )

            if success:
                tree = ET.ElementTree(ET.fromstring(output["outfiles"]["version.xml"]))
                root = tree.getroot()
                version_tree = root.find("molpro_uri:job/molpro_uri:platform/molpro_uri:version", name_space)
                if version_tree is None:
                    # some older schema
                    name_space = {"molpro_uri": "http://www.molpro.net/schema/molpro2006"}
                    version_tree = root.find("molpro_uri:job//molpro_uri:version", name_space)
                if version_tree is None:
                    return safe_version("0.0.0")
                year = version_tree.attrib["major"]
                minor = version_tree.attrib["minor"]
                molpro_version = year + "." + minor
                self.version_cache[which_prog] = safe_version(molpro_version)

        return self.version_cache[which_prog]

    def compute(self, input_data: "AtomicInput", config: "TaskConfig") -> "AtomicResult":
        """
        Run Molpro
        """
        # Check if Molpro executable is found
        self.found(raise_error=True)

        # Check Molpro version
        if parse_version(self.get_version()) < parse_version("2018.1"):
            raise TypeError("Molpro version '{}' not supported".format(self.get_version()))

        # Setup the job
        job_inputs = self.build_input(input_data, config)

        # Run Molpro
        exe_success, proc = self.execute(job_inputs)

        # Determine whether the calculation succeeded
        if exe_success:
            # If execution succeeded, collect results
            result = self.parse_output(proc["outfiles"], input_data)
            return result
        else:
            # Return UnknownError for error propagation
            return UnknownError(proc["stderr"])

    def execute(
        self,
        inputs: Dict[str, Any],
        extra_infiles: Optional[Dict[str, str]] = None,
        extra_outfiles: Optional[List[str]] = None,
        as_binary: Optional[List[str]] = None,
        extra_commands: Optional[List[str]] = None,
        scratch_name: Optional[str] = None,
        scratch_messy: bool = False,
        timeout: Optional[int] = None,
    ) -> Tuple[bool, Dict[str, Any]]:
        """
        For option documentation go look at qcengine/util.execute
        """

        # Collect all input files and update with extra_infiles
        infiles = inputs["infiles"]
        if extra_infiles is not None:
            infiles.update(extra_infiles)

        # Collect all output files and update with extra_outfiles
        outfiles = ["dispatch.out", "dispatch.xml"]
        if extra_outfiles is not None:
            outfiles.extend(extra_outfiles)

        # Replace commands with extra_commands if present
        commands = inputs["commands"]
        if extra_commands is not None:
            commands = extra_commands

        # Run the Molpro program
        exe_success, proc = execute(
            commands,
            infiles=infiles,
            outfiles=outfiles,
            as_binary=as_binary,
            scratch_name=scratch_name,
            scratch_directory=inputs["scratch_directory"],
            scratch_messy=scratch_messy,
            timeout=timeout,
        )
        return exe_success, proc

    def build_input(
        self, input_model: "AtomicInput", config: "TaskConfig", template: Optional[str] = None
    ) -> Dict[str, Any]:
        if template is None:
            input_file = []

            # Resolving keywords
            caseless_keywords = {k.lower(): v for k, v in input_model.specification.keywords.items()}
            unrestricted = False
            # Following Molpro default that ROHF is done for open-shell calculations unless unrestricted is specified
            if "reference" in caseless_keywords and caseless_keywords["reference"] == "unrestricted":
                unrestricted = True

            # Memory is in megawords per core for Molpro
            memory_mw_core = int(config.memory * (1024**3) / 8e6 / config.ncores)
            input_file.append("memory,{},M".format(memory_mw_core))
            input_file.append("")

            # Write the geom
            xyz_block = input_model.molecule.to_string(dtype="molpro", units="Bohr")
            input_file.append(xyz_block)

            # Write the basis set
            input_file.append("basis={")
            input_file.append(f"default,{input_model.specification.model.basis}")
            input_file.append("}")
            input_file.append("")

            # Determine what SCF type (restricted vs. unrestricted)
            hf_type = "RHF"
            dft_type = "RKS"
            if unrestricted:
                hf_type = "UHF"
                dft_type = "UKS"

            # Write energy call
            energy_call = []
            # If post-hf method is called then make sure to write a HF call first
            if input_model.specification.model.method.upper() in self._post_hf_methods:  # post SCF case
                energy_call.append(f"{{{hf_type}}}")
                energy_call.append("")
                energy_call.append(f"{{{input_model.specification.model.method}}}")
            # If DFT call make sure to write {rks,method}
            elif input_model.specification.model.method.upper() in self._dft_functionals:  # DFT case
                energy_call.append(f"{{{dft_type},{input_model.specification.model.method}}}")
            elif input_model.specification.model.method.upper() in self._hf_methods:  # HF case
                energy_call.append(f"{{{hf_type}}}")
            else:
                raise InputError(f"Method {input_model.specification.model.method} not implemented for Molpro.")

            # Write appropriate driver call
            if input_model.specification.driver == "energy":
                input_file.extend(energy_call)
            elif input_model.specification.driver == "gradient":
                input_file.extend(energy_call)
                input_file.append("")
                input_file.append("{force}")
            else:
                raise InputError(f"Driver {input_model.specification.driver} not implemented for Molpro.")

            input_file = "\n".join(input_file)
        else:
            # Some of the potential different template options
            # (A) ordinary build_input (need to define a base template)
            # (B) user wants to add stuff after normal template (A)
            # (C) user knows their domain language (doesn't use any QCSchema quantities)

            # # Build dictionary for substitute
            # sub_dict = {
            #     "method": input_model.model.method,
            #     "basis": input_model.model.basis,
            #     "charge": input_model.molecule.molecular_charge
            # }

            # Perform substitution to create input file
            str_template = string.Template(template)
            input_file = str_template.substitute()

        return {
            "commands": ["molpro", "dispatch.mol", "-d", ".", "-W", ".", "-n", str(config.ncores)],
            "infiles": {"dispatch.mol": input_file},
            "scratch_directory": config.scratch_directory,
            "input_result": input_model.copy(deep=True),
        }

    def parse_output(self, outfiles: Dict[str, str], input_model: "AtomicInput") -> "AtomicResult":
        tree = ET.ElementTree(ET.fromstring(outfiles["dispatch.xml"]))
        root = tree.getroot()
        # print(root.tag)

        # TODO Read information from molecule tag
        #      - cml:molecule, cml:atomArray (?)
        #      - basisSet
        #       - Be aware of symmetry. Might only be able to support if symmetry,nosym
        #      - orbitals
        #       - Be aware of symmetry. Might only be able to support if symmetry,nosym
        # NOTE: Spherical basis set ordering in Molpro (with no symmetry)
        # S -->  0
        # P --> +1, -1, 0
        # D -->  0, -2, +1, +2, -1
        # F --> +1, -1, 0, +3, -2, -3, +2
        # G -->  0, -2, +1, +4, -1, +2, -4, +3, -3
        # H --> +1, -1, +2, +3, -4, -3, +4, -5, 0, +5, -2
        # I --> +6, -2, +5, +4, -5, +2, -6, +3, -4, 0, -3, -1, +1
        properties = {}
        extras = {}
        name_space = {"molpro_uri": "http://www.molpro.net/schema/molpro-output"}

        # Molpro commands map
        molpro_map = {
            "Energy": {
                "HF": "scf_total_energy",
                "RHF": "scf_total_energy",
                "UHF": "scf_total_energy",
                "KS": "scf_total_energy",
                "RKS": "scf_total_energy",
                "UKS": "scf_total_energy",
            },
            "total energy": {
                "MP2": "mp2_total_energy",
                "CCSD": "ccsd_total_energy",
                "CCSD(T)": "ccsd_prt_pr_total_energy",
            },
            "correlation energy": {
                "MP2": "mp2_correlation_energy",
                "CCSD": "ccsd_correlation_energy",
                "CCSD(T)": "ccsd_prt_pr_correlation_energy",  # Need both CCSD(T) and Total
                "Total": "ccsd_prt_pr_correlation_energy",  # Total corresponds to CCSD(T) correlation energy
            },
            "singlet pair energy": {"MP2": "mp2_singlet_pair_energy", "CCSD": "ccsd_singlet_pair_energy"},
            "triplet pair energy": {"MP2": "mp2_triplet_pair_energy", "CCSD": "ccsd_triplet_pair_energy"},
            "Dipole moment": {
                "HF": "scf_dipole_moment",
                "RHF": "scf_dipole_moment",
                "UHF": "scf_dipole_moment",
                "KS": "scf_dipole_moment",
                "RKS": "scf_dipole_moment",
                "UKS": "scf_dipole_moment",
                "MP2": "mp2_dipole_moment",
                "CCSD": "ccsd_dipole_moment",
                "CCSD(T)": "ccsd_prt_pr_dipole_moment",
            },
        }

        # Started adding basic support for local correlation methods in Molpro
        molpro_extras_map = {
            "total energy": {
                "LMP2": "local_mp2_total_energy",
                "LCCSD": "local_ccsd_total_energy",
                "LCCSD(T0)": "local_ccsd_prt0_pr_total_energy",
                "LCCSD(T)": "local_ccsd_prt_pr_total_energy",
            },
            "correlation energy": {"LMP2": "local_mp2_correlation_energy", "LCCSD": "local_ccsd_correlation_energy"},
            "singlet pair energy": {"LMP2": "local_mp2_singlet_pair_energy", "LCCSD": "local_ccsd_singlet_pair_energy"},
            "triplet pair energy": {"LMP2": "local_mp2_triplet_pair_energy", "LCCSD": "local_ccsd_triplet_pair_energy"},
            "singles energy": {"LCCSD": "local_ccsd_singles_energy"},
            # "strong pair energy": {
            #     "LCCSD": "local_ccsd_strong_pair_energy"
            # },
            # "weak pair energy": {
            #     "LMP2": "local_mp2_weak_pair_energy"
            # }
        }

        # Molpro variables map used for quantities not found in the command map
        molpro_variable_map = {
            "_ENUC": "nuclear_repulsion_energy",
            "_DFTFUN": "scf_xc_energy",
            "_NELEC": ["calcinfo_nalpha", "calcinfo_nbeta"]
            # "_EMP2_SCS": "scs_mp2_total_energy"
        }

        # Process data in molpro_map by looping through each jobstep
        # The jobstep tag in Molpro contains output from commands (e.g. {HF}, {force})
        for jobstep in root.findall("molpro_uri:job/molpro_uri:jobstep", name_space):
            command = jobstep.attrib["command"]
            if "FORCE" in command:  # Grab gradient
                for child in jobstep.findall("molpro_uri:gradient", name_space):
                    # Stores gradient as a single list where the ordering is [1x, 1y, 1z, 2x, 2y, 2z, ...]
                    properties["gradient"] = [float(x) for x in child.text.split()]
            else:  # Grab energies and dipole moment
                for child in jobstep.findall("molpro_uri:property", name_space):
                    property_name = child.attrib["name"]
                    property_method = child.attrib["method"]
                    value = child.attrib["value"]
                    if property_name in molpro_map and property_method in molpro_map[property_name]:
                        if property_name == "Dipole moment":
                            properties[molpro_map[property_name][property_method]] = [float(x) for x in value.split()]
                        else:
                            properties[molpro_map[property_name][property_method]] = float(value)
                    elif property_name in molpro_extras_map and property_method in molpro_extras_map[property_name]:
                        extras[molpro_extras_map[property_name][property_method]] = float(value)

        # Convert triplet and singlet pair correlation energies to opposite-spin and same-spin correlation energies
        if "mp2_singlet_pair_energy" in properties and "mp2_triplet_pair_energy" in properties:
            properties["mp2_same_spin_correlation_energy"] = (2.0 / 3.0) * properties["mp2_triplet_pair_energy"]
            properties["mp2_opposite_spin_correlation_energy"] = (1.0 / 3.0) * properties[
                "mp2_triplet_pair_energy"
            ] + properties["mp2_singlet_pair_energy"]
            del properties["mp2_singlet_pair_energy"]
            del properties["mp2_triplet_pair_energy"]

        if "ccsd_singlet_pair_energy" in properties and "ccsd_triplet_pair_energy" in properties:
            properties["ccsd_same_spin_correlation_energy"] = (2.0 / 3.0) * properties["ccsd_triplet_pair_energy"]
            properties["ccsd_opposite_spin_correlation_energy"] = (1.0 / 3.0) * properties[
                "ccsd_triplet_pair_energy"
            ] + properties["ccsd_singlet_pair_energy"]
            del properties["ccsd_singlet_pair_energy"]
            del properties["ccsd_triplet_pair_energy"]

        # Process data in molpro_variable_map
        # Note: For the DFT case molecule_method is the name of the functional plus R or U in front
        molecule = root.find("molpro_uri:job/molpro_uri:molecule", name_space)
        molecule_method = molecule.attrib["method"]
        molecule_final_energy = float(molecule.attrib["energy"])  # Energy from the molecule tag in case its needed
        # Loop over each variable under the variables tag to grab additional info from molpro_variable_map
        for variable in molecule.findall("molpro_uri:variables/molpro_uri:variable", name_space):
            variable_name = variable.attrib["name"]
            if variable_name in molpro_variable_map:
                if variable_name == "_NELEC":
                    nelec = int(float(variable[0].text))
                    nunpaired = input_model.molecule.molecular_multiplicity - 1
                    nbeta = (nelec - nunpaired) // 2
                    nalpha = nelec - nbeta
                    properties[molpro_variable_map[variable_name][0]] = nalpha
                    properties[molpro_variable_map[variable_name][1]] = nbeta
                else:
                    properties[molpro_variable_map[variable_name]] = float(variable[0].text)

        # Process basis set data
        basis_set = root.find("molpro_uri:job/molpro_uri:molecule/molpro_uri:basisSet", name_space)
        nbasis = int(basis_set.attrib["length"])
        # angular_type = basis_set.attrib['angular']  # cartesian vs spherical
        properties["calcinfo_nbasis"] = nbasis

        # Grab the method from input
        method = input_model.specification.model.method.upper()

        # Determining the final energy
        # Throws an error if the energy isn't found for the method specified from the input_model.
        if method in molpro_map["total energy"].keys() and molpro_map["total energy"][method] in properties:
            final_energy = properties[molpro_map["total energy"][method]]
        elif method in molpro_map["Energy"].keys() and molpro_map["Energy"][method] in properties:
            final_energy = properties[molpro_map["Energy"][method]]
        else:
            # Back up method for determining final energy if not already present in properties
            # Use the total energy from the molecule tag if it matches the input method
            # if input_model.model.method.upper() in molecule_method:
            if method in molecule_method:
                final_energy = molecule_final_energy
                if method in self._post_hf_methods:
                    properties[molpro_map["total energy"][method]] = molecule_final_energy
                    properties[molpro_map["correlation energy"][method]] = (
                        molecule_final_energy - properties["scf_total_energy"]
                    )
                elif method in self._dft_functionals:
                    properties[molpro_map["Energy"]["KS"]] = molecule_final_energy
                elif method in self._hf_methods:
                    properties[molpro_map["Energy"][method]] = molecule_final_energy
            else:
                raise KeyError(f"Could not find {method} total energy")

        # Initialize output_data by copying over input_model.dict()
        output_data = {"input_data": input_model, "molecule": input_model.molecule}  # TODO better mol?

        # Determining return_result
        if input_model.specification.driver == "energy":
            output_data["return_result"] = final_energy
        elif input_model.specification.driver == "gradient":
            output_data["return_result"] = properties.pop("gradient")

        # Final output_data assignments needed for the AtomicResult object
        output_data["properties"] = properties
        output_data["extras"] = extras
        output_data["schema_name"] = "qcschema_atomic_result"
        output_data["stdout"] = outfiles["dispatch.out"]
        output_data["success"] = True
        output_data["provenance"] = input_model.provenance  # TODO better stamp?

        return AtomicResult(**output_data)



from typing import TYPE_CHECKING, Any, ClassVar, Dict, Union

from qcelemental.models.v2 import AtomicResult, FailedOperation, Provenance
from qcelemental.util import safe_version, which_import

from qcengine.exceptions import InputError
from qcengine.programs.model import ProgramHarness
from qcengine.units import ureg

if TYPE_CHECKING:
    from qcelemental.models.v2 import AtomicInput, FailedOperation

    from qcengine.config import TaskConfig


class MACEHarness(ProgramHarness):
    """Can be used to execute a published MACE-OFF23 model or local mace model.
    For more info on the MACE-OFF23 models see <https://doi.org/10.48550/arXiv.2312.15211>.
    The models can be found at <https://github.com/ACEsuit/mace-off>
    """

    _CACHE = {}

    _defaults: ClassVar[Dict[str, Any]] = {
        "name": "MACE",
        "scratch": False,
        "thread_safe": True,
        "thread_parallel": False,
        "node_parallel": False,
        "managed_memory": False,
    }
    version_cache: Dict[str, str] = {}

    def found(self, raise_error: bool = False) -> bool:
        return which_import(
            "mace",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via `mamba install pymace -c conda-forge`",
        )

    def get_version(self) -> str:
        self.found(raise_error=True)

        which_prog = which_import("mace")
        if which_prog not in self.version_cache:
            import mace

            self.version_cache[which_prog] = safe_version(mace.__version__)

        return self.version_cache[which_prog]

    def load_model(self, name: str):
        """Compile and cache the model to make it faster when calling many times in serial"""
        model_name = name.lower()
        if model_name in self._CACHE:
            return self._CACHE[model_name]

        import torch
        from e3nn.util import jit

        if model_name in ["small", "medium", "large"]:
            from mace.calculators.foundations_models import mace_off

            model = mace_off(model=model_name, return_raw_model=True)
        else:
            try:
                model = torch.load(name, map_location=torch.device("cpu"))
            except FileNotFoundError:
                raise InputError(
                    "The mace harness can only run local models or a MACE-OFF23 model (`small`, `medium`, `large`)"
                )
        comp_mod = jit.compile(model)
        self._CACHE[model_name] = (comp_mod, float(model.r_max), model.atomic_numbers)
        return self._CACHE[model_name]

    def compute(self, input_data: "AtomicInput", config: "TaskConfig") -> Union["AtomicResult", "FailedOperation"]:

        self.found(raise_error=True)

        import mace
        import numpy as np
        import torch
        from mace.data import AtomicData
        from mace.data.utils import AtomicNumberTable, Configuration
        from mace.tools.torch_geometric import DataLoader

        torch.set_default_dtype(torch.float64)

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        # Failure flag
        ret_data = {"success": False}

        # Build model
        method = input_data.specification.model.method

        # load the torch model which can be a MACE-OFF23 or local model
        model, r_max, atomic_numbers = self.load_model(name=method)

        z_table = AtomicNumberTable([int(z) for z in atomic_numbers])
        atomic_numbers = input_data.molecule.atomic_numbers
        pbc = (False, False, False)
        # set the cell as None and mace will automatically create a cell big enough to include all atoms
        cell = None

        config = Configuration(
            atomic_numbers=atomic_numbers,
            positions=input_data.molecule.geometry * ureg.conversion_factor("bohr", "angstrom"),
            pbc=pbc,
            cell=cell,
        )

        data_loader = DataLoader(
            dataset=[AtomicData.from_config(config, z_table=z_table, cutoff=r_max)],
            batch_size=1,
            shuffle=False,
            drop_last=False,
        )
        input_dict = next(iter(data_loader)).to_dict()
        model.to(device)
        mace_data = model(input_dict, compute_force=True)
        ret_data["properties"] = {"return_energy": mace_data["energy"] * ureg.conversion_factor("eV", "hartree")}

        if input_data.specification.driver == "energy":
            ret_data["return_result"] = ret_data["properties"]["return_energy"].detach().numpy().item()
        elif input_data.specification.driver == "gradient":
            ret_data["return_result"] = (
                np.asarray(-1.0 * mace_data["forces"] * ureg.conversion_factor("eV / angstrom", "hartree / bohr"))
                .ravel()
                .tolist()
            )

        else:
            raise InputError("MACE only supports the energy and gradient driver methods.")

        ret_data["input_data"] = input_data
        ret_data["molecule"] = input_data.molecule
        ret_data["provenance"] = Provenance(creator="mace", version=mace.__version__, routine="mace")
        ret_data["schema_name"] = "qcschema_atomic_result"
        ret_data["success"] = True

        # Form up a dict first, then sent to BaseModel to avoid repeat kwargs which don't override each other
        return AtomicResult(**ret_data)



"""
Tests the DQM compute dispatch module
"""

import warnings

import numpy as np
import pytest
import qcelemental as qcel
from qcelemental.testing import compare, compare_values

import qcengine as qcng
from qcengine.testing import checkver_and_convert, failure_engine, from_v2, schema_versions, using, uusing


@pytest.fixture(scope="function")
def input_data(request):
    if from_v2(request.node.name):
        return {
            "specification": {
                "keywords": {"coordsys": "tric", "maxiter": 100},
                "specification": {"driver": "gradient", "model": None, "keywords": {}, "program": None},
            },
            "initial_molecule": None,
        }
    else:
        return {
            "keywords": {"coordsys": "tric", "maxiter": 100, "program": None},
            "input_specification": {"driver": "gradient", "model": None, "keywords": {}},
            "initial_molecule": None,
        }


def _remove_unsupported_berny_keywords(input_data, optimizer, test_name):
    """Remove geomeTRIC-only options rejected by pyberny v0.7."""

    if optimizer == "berny":
        keywords = input_data["specification"]["keywords"] if from_v2(test_name) else input_data["keywords"]
        keywords.pop("coordsys", None)
        keywords.pop("maxiter", None)


@uusing("psi4")
@pytest.mark.parametrize("ncores", [1, 4])
@pytest.mark.parametrize(
    "optimizer",
    [
        pytest.param("geometric", marks=using("geometric")),
        pytest.param("optking", marks=using("optking")),
        pytest.param("berny", marks=using("berny")),
        pytest.param("nwchemdriver", marks=using("nwchem")),
    ],
)
def test_geometric_psi4(input_data, optimizer, ncores, schema_versions, request):
    models, retver, _ = schema_versions

    input_data["initial_molecule"] = models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True))
    if optimizer == "nwchemdriver":
        grad_program = "nwchem"
        grad_kw = {}
    else:
        grad_program = "psi4"
        grad_kw = {"scf_properties": ["wiberg_lowdin_indices"]}

    if from_v2(request.node.name):
        input_data["specification"]["specification"]["model"] = {"method": "HF", "basis": "sto-3g"}
        input_data["specification"]["specification"]["keywords"] = grad_kw
        input_data["specification"]["specification"]["program"] = grad_program
        input_data["specification"]["specification"]["extras"] = {"myqctag": "hello psi4"}
        input_data["specification"]["extras"] = {"myopttag": "hello qcengine"}
        input_data["specification"]["protocols"] = {"trajectory_results": "all"}

    else:
        input_data["input_specification"]["model"] = {"method": "HF", "basis": "sto-3g"}
        input_data["input_specification"]["keywords"] = grad_kw
        input_data["keywords"]["program"] = grad_program
        input_data["input_specification"]["extras"] = {"myqctag": "hello psi4"}
        input_data["extras"] = {"myopttag": "hello qcengine"}

    _remove_unsupported_berny_keywords(input_data, optimizer, request.node.name)
    input_data = models.OptimizationInput(**input_data)

    task_config = {
        "ncores": ncores,
    }

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", FutureWarning)
        ret = qcng.compute_procedure(
            input_data, optimizer, raise_error=True, task_config=task_config, return_version=retver
        )
    ret = checkver_and_convert(ret, request.node.name, "post")
    trajs_tgt = ret.trajectory_results if "v2" in request.node.name else ret.trajectory

    assert 10 > len(trajs_tgt) > 1

    assert pytest.approx(ret.final_molecule.measure([0, 1]), 1.0e-4) == 1.3459150737
    assert ret.provenance.creator.lower() == optimizer
    assert trajs_tgt[0].provenance.creator.lower() == grad_program

    # Note: thread passing is semi-implemented in optking and subject to other env factors
    # if optimizer == "optking" and ncores != 1:
    #     with pytest.raises(AssertionError):
    #         # Optking not passing threads to psi4
    #         assert trajs_tgt[0].provenance.nthreads == ncores
    if optimizer in ["optking", "nwchemdriver"]:
        pass
        # threading not yet implemented in NWChemDriver
    else:
        assert trajs_tgt[0].provenance.nthreads == ncores

    # Check keywords passing
    if optimizer != "nwchemdriver":
        for single in trajs_tgt:
            if "v2" in request.node.name:
                assert "scf_properties" in single.input_data.specification.keywords
            else:
                assert "scf_properties" in single.keywords
            assert (
                "WIBERG_LOWDIN_INDICES" in single.extras["qcvars"] or "WIBERG LOWDIN INDICES" in single.extras["qcvars"]
            )
            # TODO: old WIBERG qcvar used underscore; new one uses space. covering bases here but remove someday

    # Check extras passing
    if "v2" in request.node.name:
        assert (
            "myqctag" in ret.input_data.specification.specification.extras
        ), ret.input_data.specification.specification.extras.keys()
        if optimizer != "geometric":
            # geometric: below can't happen until optimizers call gradients with v2
            assert "myqctag" not in ret.trajectory_results[0].extras, "input extras wrongly present in sub-result"
        assert "myopttag" in ret.input_data.specification.extras, ret.input_data.specification.extras.keys()
        assert "myopttag" not in ret.extras, "input extras wrongly present in top-result"
    else:
        if optimizer != "optking":
            assert "myqctag" in ret.trajectory[0].extras, ret.trajectory[0].extras.keys()
            assert "myopttag" in ret.extras, ret.extras.keys()


@uusing("psi4")
@pytest.mark.parametrize(
    "optimizer",
    [
        pytest.param("geometric", marks=using("geometric")),
        pytest.param("optking", marks=using("optking")),
        pytest.param("berny", marks=using("berny")),
    ],
)
def test_geometric_local_options(input_data, schema_versions, request, optimizer):
    models, retver, _ = schema_versions

    input_data["initial_molecule"] = models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True))

    if from_v2(request.node.name):
        input_data["specification"]["specification"]["model"] = {"method": "HF", "basis": "sto-3g"}
        input_data["specification"]["specification"]["program"] = "psi4"
        input_data["specification"]["protocols"] = {"trajectory_results": "final"}
    else:
        input_data["input_specification"]["model"] = {"method": "HF", "basis": "sto-3g"}
        input_data["keywords"]["program"] = "psi4"

    _remove_unsupported_berny_keywords(input_data, optimizer, request.node.name)
    input_data = models.OptimizationInput(**input_data)

    # Set some extremely large number to test
    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, optimizer, raise_error=True, task_config={"memory": "5000"}, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    trajs_tgt = ret.trajectory_results if "v2" in request.node.name else ret.trajectory
    atin_tgt = ret.input_data.specification.specification if "v2" in request.node.name else ret.input_specification

    if optimizer == "optking":
        with pytest.raises(AssertionError):
            # Optking not passing memory to psi4
            assert pytest.approx(trajs_tgt[0].provenance.memory, 1) == 4900
    else:
        assert pytest.approx(trajs_tgt[0].provenance.memory, 1) == 4900

    # Make sure we cleaned up
    assert "_qcengine_local_config" not in atin_tgt.extras
    assert "_qcengine_local_config" not in trajs_tgt[0].extras
    if "v2" in request.node.name:
        assert "_qcengine_local_config" not in trajs_tgt[0].input_data.specification.extras


@pytest.mark.parametrize(
    "optimizer,converged",
    [
        pytest.param("geometric", "Converged!", marks=using("geometric")),
        pytest.param("optking", "Convergence check returned True", marks=using("optking")),
        pytest.param("berny", "All criteria matched", marks=using("berny")),
    ],
)
@pytest.mark.parametrize(
    "gradprog,gradmodel",
    [
        pytest.param("rdkit", {"method": "UFF", "basis": ""}, marks=using("rdkit")),
        pytest.param("psi4", {"method": "HF", "basis": "sto-3g"}, marks=using("psi4")),
    ],
)
def test_optimizer_stdout(optimizer, gradprog, gradmodel, converged, input_data, schema_versions, request):
    models, retver, _ = schema_versions

    input_data["initial_molecule"] = models.Molecule(**qcng.get_molecule("water", return_dict=True))

    if from_v2(request.node.name):
        input_data["specification"]["specification"]["model"] = gradmodel
        input_data["specification"]["specification"]["program"] = gradprog
        input_data["specification"]["specification"]["protocols"] = {"stdout": False}
        input_data["specification"]["protocols"] = {"trajectory_results": "all"}
    else:
        input_data["input_specification"]["model"] = gradmodel
        input_data["keywords"]["program"] = gradprog
        # no way to turn off gradient stdout in v1

    _remove_unsupported_berny_keywords(input_data, optimizer, request.node.name)
    input_data = models.OptimizationInput(**input_data)

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, optimizer, raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.success is True
    assert converged in ret.stdout

    trajs_tgt = ret.trajectory_results if "v2" in request.node.name else ret.trajectory
    assert ret.provenance.creator.lower() == optimizer
    assert trajs_tgt[0].provenance.creator.lower() == gradprog
    # rdkit does not have stdout, while psi4 does and is suppressed by protocol in v2 but can't be thru v1
    if from_v2(request.node.name) or gradprog == "rdkit":
        assert trajs_tgt[0].stdout is None
    else:
        assert trajs_tgt[0].stdout is not None

    if "v2" in request.node.name:
        assert ret.properties.optimization_iterations in [3, 4, 5]
        assert len(ret.trajectory_properties) == ret.properties.optimization_iterations
        assert np.allclose(ret.properties.return_gradient, ret.trajectory_results[-1].return_result)
        assert ret.trajectory_results[-1].properties.return_energy == pytest.approx(
            ret.properties.return_energy, 1.0e-4
        )
        if gradprog == "psi4":
            assert ret.properties.nuclear_repulsion_energy == pytest.approx(8.906, 1.0e-3)
            assert ret.properties.return_energy == pytest.approx(-74.96599, 1.0e-4)
        elif gradprog == "rdkit":
            assert ret.properties.nuclear_repulsion_energy == pytest.approx(8.888, 1.0e-3)
            assert ret.properties.return_energy == pytest.approx(0, 1.0e-2)  # 0 seems wrong


@pytest.mark.parametrize(
    "optimizer",
    [
        pytest.param("geometric", marks=using("geometric")),
        pytest.param("optking", marks=using("optking")),
        pytest.param("berny", marks=using("berny")),
    ],
)
@pytest.mark.parametrize(
    "gradprog,gradmodel",
    [
        pytest.param("rdkit", {"method": "UFF", "basis": ""}, marks=using("rdkit")),
        # pytest.param("psi4", {"method": "HF", "basis": "sto-3g"}, marks=using("psi4")),
    ],
)
@pytest.mark.parametrize("traj_ptcl", ["none", "final", "all", "default"])
def test_optimizer_protocols(optimizer, gradprog, gradmodel, input_data, schema_versions, request, traj_ptcl):
    models, retver, _ = schema_versions

    input_data["initial_molecule"] = models.Molecule(**qcng.get_molecule("water", return_dict=True))

    if from_v2(request.node.name):
        input_data["specification"]["specification"]["model"] = gradmodel
        input_data["specification"]["specification"]["program"] = gradprog
        if traj_ptcl != "default":
            input_data["specification"]["protocols"] = {"trajectory_results": traj_ptcl}
    else:
        input_data["input_specification"]["model"] = gradmodel
        input_data["keywords"]["program"] = gradprog
        if traj_ptcl != "default":
            input_data["protocols"] = {"trajectory": traj_ptcl}

    _remove_unsupported_berny_keywords(input_data, optimizer, request.node.name)
    input_data = models.OptimizationInput(**input_data)

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, optimizer, raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    assert ret.success is True

    trajs_tgt = ret.trajectory_results if "v2" in request.node.name else ret.trajectory
    assert ret.provenance.creator.lower() == optimizer

    if "v2" in request.node.name:
        assert len(ret.trajectory_properties) > 2

    if traj_ptcl == "none":
        assert len(trajs_tgt) == 0
    elif traj_ptcl == "final":
        assert len(trajs_tgt) == 1
    elif traj_ptcl == "all":
        assert len(trajs_tgt) > 2
    elif traj_ptcl == "default":
        if from_v2(request.node.name):
            assert len(trajs_tgt) == 0
        else:
            assert len(trajs_tgt) > 2


@uusing("psi4")
@uusing("berny")
def test_berny_failed_gradient_computation(input_data, schema_versions, request):
    models, retver, _ = schema_versions

    input_data["initial_molecule"] = models.Molecule(**qcng.get_molecule("water", return_dict=True))
    if from_v2(request.node.name):
        input_data["specification"]["specification"]["model"] = {"method": "HF", "basis": "sto-3g"}
        input_data["specification"]["specification"]["keywords"] = {"badpsi4key": "badpsi4value"}
        input_data["specification"]["specification"]["program"] = "psi4"
    else:
        input_data["input_specification"]["model"] = {"method": "HF", "basis": "sto-3g"}
        input_data["input_specification"]["keywords"] = {"badpsi4key": "badpsi4value"}
        input_data["keywords"]["program"] = "psi4"

    _remove_unsupported_berny_keywords(input_data, "berny", request.node.name)
    input_data = models.OptimizationInput(**input_data)

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "berny", raise_error=False, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post", vercheck=False)

    assert isinstance(ret, (qcel.models.v1.FailedOperation, qcel.models.v2.FailedOperation))
    assert ret.success is False
    assert ret.error.error_type == qcng.exceptions.InputError.error_type


@uusing("geometric")
@uusing("rdkit")
def test_geometric_rdkit_error(input_data, schema_versions, request):
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        input_data["specification"]["specification"]["model"] = {"method": "UFF", "basis": ""}
        input_data["specification"]["specification"]["program"] = "rdkit"
        im = qcng.get_molecule("water", return_dict=True)
        im.pop("connectivity")
        input_data["initial_molecule"] = models.Molecule(**im).model_copy()
    else:
        input_data["input_specification"]["model"] = {"method": "UFF", "basis": ""}
        input_data["keywords"]["program"] = "rdkit"
        input_data["initial_molecule"] = im = models.Molecule(**qcng.get_molecule("water", return_dict=True)).copy(
            exclude={"connectivity_"}
        )

    input_data = models.OptimizationInput(**input_data)

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "geometric", return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post", vercheck=False)

    assert ret.success is False
    assert isinstance(ret.error.error_message, str)
    assert isinstance(ret, (qcel.models.v1.FailedOperation, qcel.models.v2.FailedOperation))


@pytest.mark.parametrize(
    "optimizer",
    [
        pytest.param("geometric", marks=using("geometric")),
        pytest.param("optking", marks=using("optking")),
    ],
)
def test_geometric_maxiter_failure(optimizer, failure_engine, input_data, schema_versions, request):
    models, retver, _ = schema_versions

    failure_engine.iter_modes = ["pass"] * 20

    input_data["initial_molecule"] = {
        "symbols": ["He", "He"],
        "geometry": [0, 0, 0, 0, 0, failure_engine.start_distance],
    }

    if from_v2(request.node.name):
        input_data["specification"]["specification"]["model"] = {"method": "something"}
        input_data["specification"]["specification"]["program"] = failure_engine.name
        if optimizer == "geometric":
            input_data["specification"]["keywords"].update({"coordsys": "cart", "maxiter": 1})
        elif optimizer == "optking":
            input_data["specification"]["keywords"].update({"geom_maxiter": 1})
            input_data["specification"]["keywords"].pop("maxiter")
    else:
        input_data["input_specification"]["model"] = {"method": "something"}
        input_data["keywords"]["program"] = failure_engine.name
        if optimizer == "optking":
            input_data["keywords"].update({"geom_maxiter": 1})
            input_data["keywords"].pop("maxiter", None)
        else:
            input_data["keywords"].update({"coordsys": "cart", "maxiter": 1})

    input_data = models.OptimizationInput(**input_data)

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, optimizer, raise_error=False, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post", vercheck=False)

    assert ret.success is False
    assert isinstance(ret, (qcel.models.v1.FailedOperation, qcel.models.v2.FailedOperation))
    assert ret.error.extras and "failed_result" in ret.error.extras

    assert not hasattr(ret, "stdout")
    if "v2" in request.node.name:
        assert ret.input_data.schema_name == "qcschema_optimization_input"

    retgrad = [0.0, 0.0, -1.0, 0.0, 0.0, 1.0]
    if optimizer == "geometric":
        # There's more iterations than wanted (trajectory[0] != trajectory[-1]) with gradients
        #   [0.0, 0.0, -0.6220, 0.0, 0.0, 0.6220] (v1.1) or [0,0,-0.0875 ... (v1.0.2), but this
        #   test is focused on structure of FailedOp.
        assert r"failed to converge" in ret.error.error_message.lower(), ret.error.error_message
        assert (
            "Maximum iterations reached (1); increase --maxiter for more" in ret.error.extras["failed_result"]["stdout"]
        )
        assert "Job type: Energy minimization" in ret.error.extras["failed_result"]["stdout"]
        final_result = ret.error.extras["failed_result"]["trajectory"][0]["return_result"]
        assert compare_values(
            retgrad,
            final_result,
            "fake gradient",
            atol=1.0e-4,
        )
        assert ret.error.extras["failed_result"]["provenance"]["creator"] == "geomeTRIC"
        assert compare_values(
            retgrad, ret.error.extras["failed_result"]["trajectory"][0]["return_result"], "fake gradient traj"
        )  # note v1 field name
    elif optimizer == "optking":
        import optking as ok

        assert r"maximum number of steps exceeded: 1" in ret.error.error_message.lower(), ret.error.error_message
        assert "Dumping history: Warning last point not converged" in ret.error.extras["failed_result"]["stdout"]
        assert ret.error.extras["failed_result"]["provenance"]["creator"] == "optking"
        if int(ok.__version__.split(".")[1]) < 5:
            assert compare_values(
                retgrad, ret.error.extras["failed_result"]["trajectory"][0]["return_result"], "fake gradient traj"
            )  # note v1 field name
        else:
            assert compare_values(
                retgrad,
                ret.error.extras["failed_result"]["properties"]["return_gradient"],
                "fake gradient",
                atol=1.0e-6,
            )
            assert ret.error.extras["failed_result"]["properties"]["optimization_iterations"] == 1
            assert compare_values(
                retgrad,
                ret.error.extras["failed_result"]["trajectory_results"][0]["return_result"],
                "fake gradient traj",
            )  # note v2 field name


@uusing("rdkit")
@pytest.mark.parametrize(
    "optimizer",
    [
        pytest.param("geometric", marks=using("geometric")),
        pytest.param("optking", marks=using("optking")),
        pytest.param("berny", marks=using("berny")),
        pytest.param("nwchemdriver", marks=using("nwchem")),
    ],
)
def test_optimization_protocols(optimizer, input_data, schema_versions, request):
    models, retver, _ = schema_versions

    if optimizer == "nwchemdriver":
        grad_program = "nwchem"
        grad_model = {"method": "HF", "basis": "sto-3g"}
    else:
        grad_program = "rdkit"
        grad_model = {"method": "UFF"}

    input_data["initial_molecule"] = models.Molecule(**qcng.get_molecule("water", return_dict=True))
    if from_v2(request.node.name):
        input_data["specification"]["specification"]["model"] = grad_model
        input_data["specification"]["specification"]["program"] = grad_program
        input_data["specification"]["protocols"] = {"trajectory_results": "initial_and_final"}
    else:
        input_data["input_specification"]["model"] = grad_model
        input_data["keywords"]["program"] = grad_program
        input_data["protocols"] = {"trajectory": "initial_and_final"}

    _remove_unsupported_berny_keywords(input_data, optimizer, request.node.name)
    input_data = models.OptimizationInput(**input_data)

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, optimizer, raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    trajs_tgt = ret.trajectory_results if "v2" in request.node.name else ret.trajectory
    initmol_tgt = ret.input_data.initial_molecule if "v2" in request.node.name else ret.initial_molecule

    assert ret.success, ret.error.error_message
    assert len(trajs_tgt) == 2
    assert (
        pytest.approx(initmol_tgt.nuclear_repulsion_energy(), 1.0e-4)
        == trajs_tgt[0].molecule.nuclear_repulsion_energy()
    )
    if optimizer != "nwchemdriver":
        # in internal-opt mode, fix_/orient is not called, so the hash will be different
        assert initmol_tgt.get_hash() == trajs_tgt[0].molecule.get_hash()
    assert (
        pytest.approx(ret.final_molecule.nuclear_repulsion_energy(), 1.0e-4)
        == trajs_tgt[1].molecule.nuclear_repulsion_energy()
    )
    if optimizer not in ["nwchemdriver", "optking"]:
        # optking would pass if: optking/optimize.py: #computer.update_geometry(o_molsys.geom)
        assert ret.final_molecule.get_hash() == trajs_tgt[1].molecule.get_hash()


@uusing("geometric")
def test_geometric_retries(failure_engine, input_data, schema_versions, request):
    import geometric

    tric_ver = geometric.__version__
    models, retver, _ = schema_versions

    failure_engine.iter_modes = ["random_error", "pass", "random_error", "random_error", "pass"]  # Iter 1  # Iter 2
    failure_engine.iter_modes.extend(["pass"] * 20)

    input_data["initial_molecule"] = {
        "symbols": ["He", "He"],
        "geometry": [0, 0, 0, 0, 0, failure_engine.start_distance],
    }

    if from_v2(request.node.name):
        input_data["specification"]["specification"]["model"] = {"method": "something"}
        input_data["specification"]["specification"]["program"] = failure_engine.name
        input_data["specification"]["keywords"][
            "coordsys"
        ] = "cart"  # needed by geometric v1.0 to play nicely with failure_engine
        input_data["specification"]["protocols"] = {"trajectory_results": "all"}
    else:
        input_data["input_specification"]["model"] = {"method": "something"}
        input_data["keywords"]["program"] = failure_engine.name
        input_data["keywords"]["coordsys"] = "cart"  # needed by geometric v1.0 to play nicely with failure_engine

    input_data = models.OptimizationInput(**input_data)

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "geometric", task_config={"ncores": 13}, raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    trajs_tgt = ret.trajectory_results if "v2" in request.node.name else ret.trajectory

    assert ret.success is True
    assert trajs_tgt[0].provenance.retries == 1
    assert trajs_tgt[0].provenance.ncores == 13
    assert trajs_tgt[1].provenance.retries == 2
    assert trajs_tgt[1].provenance.ncores == 13
    assert "retries" not in trajs_tgt[2].provenance.model_dump()

    # Ensure we still fail
    failure_engine.iter_modes = ["random_error", "pass", "random_error", "random_error", "pass"]  # Iter 1  # Iter 2

    ret = qcng.compute(input_data, "geometric", task_config={"ncores": 13, "retries": 1}, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post", vercheck=False)

    assert ret.success is False
    assert ret.error.extras and "failed_result" in ret.error.extras
    if "v2" in request.node.name:
        assert compare(["He", "He"], ret.input_data.initial_molecule.symbols, "symbols")
    else:
        assert compare(["He", "He"], ret.input_data["initial_molecule"]["symbols"], "symbols")
    failed_result = ret.error.extras["failed_result"]
    failed_traj = failed_result.get("trajectory", failed_result.get("trajectory_results"))
    assert failed_traj[0]["provenance"]["retries"] == 1
    if tric_ver in ["1.1", "1.1.1"]:
        # bad! temp until https://github.com/leeping/geomeTRIC/pull/222 available
        assert len(failed_traj) == 1
    else:
        assert len(failed_traj) == 2


@uusing("geometric")
@pytest.mark.parametrize(
    "program, model, bench",
    [
        pytest.param(
            "rdkit", {"method": "UFF"}, [1.87130923886072, 2.959448636243545, 104.5099642579023], marks=using("rdkit")
        ),
        pytest.param(
            "rdkit",
            {"method": "mmff94"},
            [1.8310842343589573, 2.884612338953529, 103.93822919865106],
            marks=using("rdkit"),
        ),
        pytest.param(
            "rdkit",
            {"method": "MMFF94s"},
            [1.8310842343589573, 2.884612338953529, 103.93822919865106],
            marks=using("rdkit"),
        ),
        pytest.param(
            "torchani",
            {"method": "ANI1x"},
            [1.82581873750194, 2.866376526793269, 103.4332610730292],
            marks=using("torchani"),
        ),
        pytest.param(
            "mopac",
            {"method": "PM6"},
            [1.793052302291527, 2.893333237502448, 107.57254391453196],
            marks=using("mopac"),
        ),
        pytest.param(
            "openmm",
            {"method": "openff-1.0.0", "basis": "smirnoff"},
            [1.8344994291796748, 3.010099477501204, 110.25177977849998],
            marks=using("openmm"),
        ),
        pytest.param(
            "openmm",
            {"method": "openff_unconstrained-1.0.0", "basis": "smirnoff"},
            [1.8344994291195869, 3.0100994772976124, 110.25259556886984],
            marks=using("openmm"),
        ),
        pytest.param(
            "openmm",
            {"method": "smirnoff99Frosst-1.1.0", "basis": "smirnoff"},
            [1.814137087600702, 3.025566213038376, 112.9999999990053],
            marks=using("openmm"),
        ),
        pytest.param(
            "qcore",
            {"method": "GFN1"},
            [1.8104763949897031, 2.9132449420655213, 107.13403040879244],
            marks=using("qcore"),
        ),
    ],
)
def test_geometric_generic(input_data, program, model, bench, schema_versions, request):
    models, retver, _ = schema_versions

    input_data["initial_molecule"] = models.Molecule(**qcng.get_molecule("water", return_dict=True))
    if from_v2(request.node.name):
        input_data["specification"]["specification"]["model"] = model
        input_data["specification"]["specification"]["program"] = program
        input_data["specification"]["specification"]["extras"] = {
            "_secret_tags": {"mysecret_tag": "data1"}  # pragma: allowlist secret
        }
        input_data["specification"]["protocols"] = {"trajectory_results": "final"}
    else:
        input_data["input_specification"]["model"] = model
        input_data["keywords"]["program"] = program
        input_data["input_specification"]["extras"] = {
            "_secret_tags": {"mysecret_tag": "data1"}  # pragma: allowlist secret
        }

    input_data = checkver_and_convert(input_data, request.node.name, "pre", cast_dict_as="OptimizationInput")
    ret = qcng.compute(input_data, "geometric", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    trajs_tgt = ret.trajectory_results if "v2" in request.node.name else ret.trajectory

    assert ret.success is True
    assert "Converged!" in ret.stdout

    r01, r02, r12, a102 = ret.final_molecule.measure([[0, 1], [0, 2], [1, 2], [1, 0, 2]])

    assert pytest.approx(r01, 1.0e-4) == bench[0]
    assert pytest.approx(r02, 1.0e-4) == bench[0]
    assert pytest.approx(r12, 5.0e-4) == bench[1]
    assert pytest.approx(a102, 5.0e-4) == bench[2]

    assert "_secret_tags" in trajs_tgt[0].extras
    assert "data1" == trajs_tgt[0].extras["_secret_tags"]["mysecret_tag"], trajs_tgt[0].extras["_secret_tags"]
    # the _secret_tags shows up in Res.extras, not in Res.input_data.extras b/c geometric running v1 internally


@uusing("nwchem")
@pytest.mark.parametrize("linopt", [0, 1])
def test_nwchem_relax(linopt, schema_versions, request):
    models, retver, _ = schema_versions

    # Make the input file
    if from_v2(request.node.name):
        input_data = {
            "specification": {
                "specification": {
                    "model": {"method": "HF", "basis": "sto-3g"},
                    "keywords": {"set__driver:linopt": linopt},
                    "driver": "gradient",
                },
                "protocols": {"trajectory_results": "all"},
            },
            "initial_molecule": models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True)),
        }
    else:
        input_data = {
            "input_specification": {
                "model": {"method": "HF", "basis": "sto-3g"},
                "keywords": {"set__driver:linopt": linopt},
            },
            "initial_molecule": models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True)),
        }
    input_data = models.OptimizationInput(**input_data)

    # Run the relaxation
    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "nwchemdriver", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    trajs_tgt = ret.trajectory_results if "v2" in request.node.name else ret.trajectory

    assert 10 > len(trajs_tgt) > 1

    assert pytest.approx(ret.final_molecule.measure([0, 1]), 1.0e-4) == 1.3459150737


@uusing("nwchem")
def test_nwchem_restart(tmpdir, schema_versions, request):
    models, retver, _ = schema_versions

    # Make the input file
    if from_v2(request.node.name):
        input_data = {
            "specification": {
                "specification": {
                    "model": {"method": "HF", "basis": "sto-3g"},
                    "keywords": {"driver__maxiter": 2, "set__driver:linopt": 0},
                    "driver": "gradient",  # newly req'd by model in v2
                    "extras": {"allow_restarts": True},
                },
            },
            "initial_molecule": models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True)),
        }
    else:
        input_data = {
            "input_specification": {
                "model": {"method": "HF", "basis": "sto-3g"},
                "keywords": {"driver__maxiter": 2, "set__driver:linopt": 0},
                "extras": {"allow_restarts": True},
            },
            "initial_molecule": models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True)),
        }
    input_data = models.OptimizationInput(**input_data)

    # Run an initial step, which should not converge
    local_opts = {"scratch_messy": True, "scratch_directory": str(tmpdir)}

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "nwchemdriver", task_config=local_opts, raise_error=False, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post", vercheck=False)
    assert not ret.success

    # Run it again, which should converge
    new_ret = qcng.compute(input_data, "nwchemdriver", task_config=local_opts, raise_error=True, return_version=retver)
    new_ret = checkver_and_convert(new_ret, request.node.name, "post")
    assert new_ret.success


@uusing("rdkit")
@uusing("torsiondrive")
@pytest.mark.parametrize("scan_ptcl", ["none", "all", "lowest"])
def test_torsiondrive_generic(schema_versions, request, scan_ptcl):
    models, retver, _ = schema_versions

    if from_v2(request.node.name):
        input_data = models.TorsionDriveInput(
            initial_molecule=[models.Molecule(**qcng.get_molecule("ethane", return_dict=True))] * 2,
            specification=models.TorsionDriveSpecification(
                keywords=models.TorsionDriveKeywords(dihedrals=[(2, 0, 1, 5)], grid_spacing=[180]),
                protocols=models.TorsionDriveProtocols(scan_results=scan_ptcl),
                specification=models.OptimizationSpecification(
                    protocols=models.OptimizationProtocols(trajectory_results="all"),
                    program="geomeTRIC",
                    keywords={
                        "coordsys": "hdlc",
                        "maxiter": 500,
                    },
                    specification=models.AtomicSpecification(
                        program="rdkit", driver=models.DriverEnum.gradient, model=models.Model(method="UFF", basis=None)
                    ),
                ),
            ),
        )
    else:
        input_data = models.TorsionDriveInput(
            keywords=models.TDKeywords(dihedrals=[(2, 0, 1, 5)], grid_spacing=[180]),
            input_specification=models.QCInputSpecification(
                driver=models.DriverEnum.gradient, model=models.Model(method="UFF", basis=None)
            ),
            initial_molecule=[models.Molecule(**qcng.get_molecule("ethane", return_dict=True))] * 2,
            optimization_spec=models.OptimizationSpecification(
                procedure="geomeTRIC",
                keywords={
                    "coordsys": "hdlc",
                    "maxiter": 500,
                    "program": "rdkit",
                },
            ),
        )

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "torsiondrive", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    if "_v2" not in request.node.name:
        assert ret.error is None
    assert ret.success

    expected_grid_ids = {"180", "0"}
    opthist_tgt = ret.scan_results if "v2" in request.node.name else ret.optimization_history

    if not (from_v2(request.node.name) and scan_ptcl == "none"):
        assert {*opthist_tgt} == expected_grid_ids

    if "v2" in request.node.name:
        assert {*ret.scan_properties} == expected_grid_ids
    else:
        assert {*ret.final_energies} == expected_grid_ids
    assert {*ret.final_molecules} == expected_grid_ids

    assert (
        pytest.approx(ret.final_molecules["180"].measure([2, 0, 1, 5]), abs=1.0e-2) == 180.0
        or pytest.approx(ret.final_molecules["180"].measure([2, 0, 1, 5]), abs=1.0e-2) == -180.0
    )
    assert pytest.approx(ret.final_molecules["0"].measure([2, 0, 1, 5]), abs=1.0e-2) == 0.0

    assert ret.provenance.creator.lower() == "torsiondrive"

    if from_v2(request.node.name):
        # properly "to_v1" should be always `== 4` but data lost in v2 by default
        if scan_ptcl == "none":
            assert not opthist_tgt
        else:
            assert len(opthist_tgt["180"]) == {"lowest": 1, "all": 4}[scan_ptcl]
    else:
        assert len(opthist_tgt["180"]) == 4

    if not (from_v2(request.node.name) and scan_ptcl == "none"):
        assert opthist_tgt["180"][0].provenance.creator.lower() == "geometric"
    if "v2" in request.node.name:
        if scan_ptcl != "none":
            assert opthist_tgt["180"][0].trajectory_results[0].provenance.creator.lower() == "rdkit"
            assert opthist_tgt["180"][0].trajectory_results[0].schema_version == 2
    else:
        if not ("to_v1" in request.node.name and scan_ptcl == "none"):
            assert opthist_tgt["180"][0].trajectory[0].provenance.creator.lower() == "rdkit"
            assert opthist_tgt["180"][0].trajectory[0].schema_version == 1

    assert ret.stdout == "All optimizations converged at lowest energy. Job Finished!\n"


@uusing("mace")
@uusing("torsiondrive")
def test_torsiondrive_extra_constraints(schema_versions, request):
    models, retver, _ = schema_versions

    keywords = {
        "coordsys": "dlc",
        "constraints": {
            "set": [
                {
                    "type": "dihedral",  # hold a dihedral through the other C-C bond fixed
                    "indices": (0, 1, 2, 10),
                    "value": 0.0,
                }
            ]
        },
    }

    if from_v2(request.node.name):
        input_data = models.TorsionDriveInput(
            initial_molecule=[models.Molecule(**qcng.get_molecule("propane", return_dict=True))],
            specification=models.TorsionDriveSpecification(
                keywords=models.TorsionDriveKeywords(dihedrals=[(3, 0, 1, 2)], grid_spacing=[180]),
                specification=models.OptimizationSpecification(
                    program="geomeTRIC",
                    keywords=keywords,
                    specification=models.AtomicSpecification(
                        # use mace as it does not have convergence issues like UFF
                        program="mace",
                        driver=models.DriverEnum.gradient,
                        model=models.Model(method="small", basis=None),
                    ),
                    protocols=models.OptimizationProtocols(trajectory_results="all"),
                ),
                protocols=models.TorsionDriveProtocols(scan_results="all"),
            ),
        )
    else:
        input_data = models.TorsionDriveInput(
            keywords=models.TDKeywords(dihedrals=[(3, 0, 1, 2)], grid_spacing=[180]),
            input_specification=models.QCInputSpecification(
                driver=models.DriverEnum.gradient, model=models.Model(method="small", basis=None)
            ),
            initial_molecule=[models.Molecule(**qcng.get_molecule("propane", return_dict=True))],
            optimization_spec=models.OptimizationSpecification(
                procedure="geomeTRIC",
                keywords={
                    **keywords,
                    # use mace as it does not have convergence issues like UFF
                    "program": "mace",
                },
            ),
        )

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, "torsiondrive", raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    if "_v2" not in request.node.name:
        assert ret.error is None
    assert ret.success

    expected_grid_ids = {"180", "0"}
    opthist_tgt = ret.scan_results if "v2" in request.node.name else ret.optimization_history

    assert {*opthist_tgt} == expected_grid_ids
    if "v2" in request.node.name:
        assert {*ret.scan_properties} == expected_grid_ids
    else:
        assert {*ret.final_energies} == expected_grid_ids
    assert {*ret.final_molecules} == expected_grid_ids
    if "v2" in request.node.name:
        assert ret.properties.calcinfo_ngrid == 2

    assert (
        pytest.approx(ret.final_molecules["180"].measure([3, 0, 1, 2]), abs=1.0e-2) == 180.0
        or pytest.approx(ret.final_molecules["180"].measure([3, 0, 1, 2]), abs=1.0e-2) == -180.0
    )
    assert pytest.approx(ret.final_molecules["180"].measure([0, 1, 2, 10]), abs=1.0e-2) == 0.0
    assert pytest.approx(ret.final_molecules["0"].measure([3, 0, 1, 2]), abs=1.0e-2) == 0.0

    assert ret.provenance.creator.lower() == "torsiondrive"
    assert opthist_tgt["180"][0].provenance.creator.lower() == "geometric"
    if "v2" in request.node.name:
        assert opthist_tgt["180"][0].trajectory_results[0].provenance.creator.lower() == "mace"
    else:
        assert opthist_tgt["180"][0].trajectory[0].provenance.creator.lower() == "mace"

    assert "Using MACE-OFF23 MODEL for MACECalculator" in ret.stdout
    assert "All optimizations converged at lowest energy. Job Finished!\n" in ret.stdout


@uusing("mrchem")
@pytest.mark.parametrize(
    "optimizer",
    [
        pytest.param("geometric", marks=using("geometric")),
        pytest.param("optking", marks=using("optking")),
        pytest.param("berny", marks=using("berny")),
    ],
)
def test_optimization_mrchem(input_data, optimizer, schema_versions, request):
    models, retver, _ = schema_versions

    input_data["initial_molecule"] = models.Molecule(**qcng.get_molecule("hydrogen", return_dict=True))
    if from_v2(request.node.name):
        input_data["specification"]["specification"]["model"] = {"method": "HF"}
        input_data["specification"]["specification"]["keywords"] = {"world_prec": 1.0e-4}
        input_data["specification"]["specification"]["program"] = "mrchem"
        input_data["specification"]["protocols"] = {"trajectory_results": "final"}  # to test provenance
    else:
        input_data["input_specification"]["model"] = {"method": "HF"}
        input_data["input_specification"]["keywords"] = {"world_prec": 1.0e-4}
        input_data["keywords"]["program"] = "mrchem"

    _remove_unsupported_berny_keywords(input_data, optimizer, request.node.name)
    input_data = models.OptimizationInput(**input_data)

    input_data = checkver_and_convert(input_data, request.node.name, "pre")
    ret = qcng.compute(input_data, optimizer, raise_error=True, return_version=retver)
    ret = checkver_and_convert(ret, request.node.name, "post")

    trajs_tgt = ret.trajectory_results if "v2" in request.node.name else ret.trajectory  # for looking at grad jobs
    props_tgt = ret.trajectory_properties if "v2" in request.node.name else ret.energies  # for counting opt iterations

    assert 10 > len(props_tgt) > 1
    if "v2" in request.node.name:
        assert 10 > ret.properties.optimization_iterations > 1

    assert pytest.approx(ret.final_molecule.measure([0, 1]), 1.0e-3) == 1.3860734486984705
    assert ret.provenance.creator.lower() == optimizer
    assert trajs_tgt[0].provenance.creator.lower() == "mrchem"



from types import SimpleNamespace

import numpy as np
import pytest
from qcelemental import constants
from qcelemental.models.v2 import AtomicInput
from qcelemental.util import safe_version

import qcengine as qcng
from qcengine.exceptions import InputError
from qcengine.programs.jaguar import JaguarHarness
from qcengine.testing import uusing


def _input(driver="energy", method="hf", basis="sto-3g", keywords=None, extras=None, charge=0, multiplicity=1):
    return AtomicInput(
        molecule={
            "symbols": ["H", "H"],
            "geometry": [0.0, 0.0, -0.7, 0.0, 0.0, 0.7],
            "molecular_charge": charge,
            "molecular_multiplicity": multiplicity,
        },
        specification={
            "driver": driver,
            "model": {"method": method, "basis": basis},
            "keywords": keywords or {},
            "extras": extras or {},
            "protocols": {"native_files": "all"},
        },
    )


def test_registered():
    assert "jaguar" in qcng.list_all_programs()
    assert isinstance(qcng.get_program("jaguar", check=False), JaguarHarness)


@uusing("jaguar")
def test_versions():
    from schrodinger.application.jaguar.constants import VERSION
    from schrodinger.infra import mm

    encoded_version = int(mm.mmfile_get_product_version("jaguar"))
    major_minor, release = divmod(encoded_version, 1000)
    major, minor = divmod(major_minor, 10)
    harness = qcng.get_program("jaguar")

    assert harness.get_version() == safe_version(f"{major}.{minor}.{release:03d}")
    assert harness.get_suite_version() == safe_version(VERSION)


def test_property_mapping():
    dipole = SimpleNamespace(x=1.0, y=2.0, z=3.0)
    results = SimpleNamespace(
        energy=-1.1,
        nuclear_repulsion=0.7,
        scf_energy=-1.0,
        energy_one_electron=-2.0,
        energy_two_electron=0.3,
        rimp2_ss_energy=-0.01,
        rimp2_os_energy=-0.04,
        rimp2_corr_energy=-0.05,
        rimp2_energy=-1.05,
        dipole_qm=dipole,
    )
    output = SimpleNamespace(
        last_results=results,
        nbasis=5,
        num_occ_orbs_alpha=None,
        num_occ_orbs_beta=None,
        num_occ_orbs=1,
    )

    properties = JaguarHarness._optional_properties(output, _input())

    assert properties["calcinfo_natom"] == 2
    assert properties["calcinfo_nbasis"] == 5
    assert properties["calcinfo_nmo"] == 5
    assert properties["calcinfo_nalpha"] == properties["calcinfo_nbeta"] == 1
    assert properties["return_energy"] == -1.1
    assert properties["scf_total_energy"] == -1.0
    assert properties["mp2_total_energy"] == -1.05
    assert np.allclose(
        properties["scf_dipole_moment"],
        np.array([1.0, 2.0, 3.0]) / constants.dipmom_au2debye,
    )


def test_rohf_electron_counts():
    output = SimpleNamespace(
        nelectron=7,
        nbasis=5,
        num_occ_orbs_alpha=4,
        num_occ_orbs_beta=4,
        last_results=SimpleNamespace(energy=-1.1),
    )
    input_model = _input(charge=1, multiplicity=2)

    properties = JaguarHarness._optional_properties(output, input_model)

    assert properties["calcinfo_nalpha"] == 4
    assert properties["calcinfo_nbeta"] == 3


@pytest.mark.parametrize(
    "driver, derivative_label",
    [("energy", None), ("gradient", "GRADIENT"), ("hessian", "HESSIAN")],
)
def test_hf_qcvars(driver, derivative_label):
    properties = {"return_energy": -1.0, "scf_total_energy": -1.0}
    if derivative_label:
        properties[f"return_{derivative_label.lower()}"] = np.zeros((2, 3))

    qcvars = JaguarHarness._qcvars(properties, driver, "hf")

    assert qcvars["HF TOTAL ENERGY"] == -1.0
    assert qcvars["SCF TOTAL ENERGY"] == -1.0
    assert qcvars["CURRENT REFERENCE ENERGY"] == -1.0
    assert qcvars["CURRENT ENERGY"] == -1.0
    assert "CURRENT CORRELATION ENERGY" not in qcvars
    if derivative_label:
        assert f"HF TOTAL {derivative_label}" in qcvars
        assert f"CURRENT {derivative_label}" in qcvars


def test_rejects_unsupported_driver():
    with pytest.raises(InputError, match="not implemented"):
        JaguarHarness._validate_input(_input(driver="properties"))


def test_rejects_invalid_guess_input():
    input_model = _input(extras={"jaguar": {"guess_input": 42}})
    with pytest.raises(InputError, match="guess_input must be a non-empty string"):
        JaguarHarness._validate_input(input_model)


@uusing("jaguar")
def test_input_mapping(tmp_path):
    input_model = _input(
        driver="gradient",
        method="b3lyp-d3",
        basis="6-31g**",
        keywords={"maxit": 99, "dftname": "bad", "basis": "bad", "igeopt": 2, "isymm": 8},
    )

    jaguar_input = JaguarHarness._build_jaguar_input(input_model, str(tmp_path / "dispatch"))

    assert jaguar_input.getValue("dftname").lower() == "b3lyp-d3"
    assert jaguar_input.getValue("basis").lower() == "6-31g**"
    assert jaguar_input.getValue("maxit") == 99
    assert jaguar_input.getValue("igeopt") == -1
    assert jaguar_input.getValue("ifreq") == 0
    assert jaguar_input.getValue("isymm") == 0
    assert jaguar_input.getValue("molchg") == 0
    assert jaguar_input.getValue("multip") == 1
    assert np.allclose(
        [[atom.x, atom.y, atom.z] for atom in jaguar_input.getStructure().atom],
        np.asarray(input_model.molecule.geometry) * constants.bohr2angstroms,
    )


@uusing("jaguar")
def test_guess_input_mapping(tmp_path):
    guess_text = """&gen
maxit=7
basis=6-31g
&
&zmat
H1 0.0 0.0 -1.0
H2 0.0 0.0  1.0
&
&guess basgss=sto-3g numd=1
    1 Orbital Energy -0.500000 Occupation 1.000000 Symmetry A
  1.000000 0.000000
&
"""
    input_model = _input(
        driver="gradient",
        keywords={"maxit": 99},
        extras={"jaguar": {"guess_input": guess_text}},
    )
    job_base = tmp_path / "guess"

    JaguarHarness._validate_input(input_model)
    jaguar_input = JaguarHarness._build_jaguar_input(input_model, str(job_base))

    assert jaguar_input.getValue("maxit") == 99
    assert jaguar_input.getValue("igeopt") == -1
    assert jaguar_input.sectionDefined("guess")
    assert "Orbital Energy -0.500000" in jaguar_input.getSectionText("guess")

    jaguar_input.save()
    output_text = job_base.with_suffix(".in").read_text()
    assert "&guess" in output_text
    assert "maxit=99" in output_text.replace(" ", "")
    assert "maxit=7" not in output_text.replace(" ", "")


@uusing("jaguar")
def test_ghost_atom_mapping(tmp_path):
    input_model = AtomicInput(
        molecule={
            "symbols": ["H", "He"],
            "geometry": [0.0, 0.0, 0.0, 0.0, 0.0, 3.0],
            "real": [True, False],
            "molecular_charge": 0,
            "molecular_multiplicity": 2,
        },
        specification={
            "driver": "energy",
            "model": {"method": "hf", "basis": "sto-3g"},
        },
    )
    job_base = tmp_path / "ghost"

    JaguarHarness._validate_input(input_model)
    jaguar_input = JaguarHarness._build_jaguar_input(input_model, str(job_base))
    jaguar_atoms = list(jaguar_input.getStructure().atom)

    assert [atom.atomic_number for atom in jaguar_atoms] == [1, 2]

    jaguar_input.save()
    input_text = job_base.with_suffix(".in").read_text()
    assert "H1" in input_text
    assert "He2@" in input_text


@uusing("jaguar")
@pytest.mark.parametrize(
    "driver, shape",
    [
        ("energy", ()),
        ("gradient", (2, 3)),
        ("hessian", (6, 6)),
    ],
)
def test_compute(driver, shape, tmp_path):
    result = qcng.compute(
        _input(driver=driver),
        "jaguar",
        raise_error=True,
        task_config={"ncores": 1, "scratch_directory": str(tmp_path)},
        return_version=2,
    )

    harness = qcng.get_program("jaguar")
    assert result.success
    assert np.asarray(result.return_result).shape == shape
    assert result.properties.return_energy is not None
    assert result.provenance.version == harness.get_version()
    assert set(result.extras["jaguar"]) == {"suite_version", "point_group"}
    assert result.extras["jaguar"]["suite_version"] == harness.get_suite_version()
    assert "Jaguar version" in result.stdout
    assert "&gen" in result.native_files["input"]
    generated_input = result.native_files["dispatch.01.in"]
    assert "&gen" in generated_input
    if driver == "hessian":
        assert result.properties.return_hessian.shape == shape
        assert "&hess" in generated_input
